# -*- coding: utf-8 -*-

"""Spike detection."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from six import string_types
from six.moves import range, zip


#------------------------------------------------------------------------------
# Thresholder
#------------------------------------------------------------------------------

def compute_threshold(arr, single_threshold=True, std_factor=None):
    """Compute the threshold(s) of filtered traces.

    Parameters
    ----------

    arr : ndarray
        Filtered traces, shape `(n_samples, n_channels)`.
    single_threshold : bool
        Whether there should be a unique threshold for all channels, or
        one threshold per channel.
    std_factor : float or 2-tuple
        The threshold in unit of signal std. Two values can be specified
        for multiple thresholds (weak and strong).

    Returns
    -------

    thresholds : ndarray
        A `(2,)` or `(2, n_channels)` array with the thresholds.

    """
    assert arr.ndim == 2
    ns, nc = arr.shape

    assert std_factor is not None
    if isinstance(std_factor, (int, float)):
        std_factor = (std_factor, std_factor)
    assert isinstance(std_factor, (tuple, list))
    assert len(std_factor) == 2
    std_factor = np.array(std_factor)

    if not single_threshold:
        std_factor = std_factor[:, None]

    # Get the median of all samples in all excerpts, on all channels.
    if single_threshold:
        median = np.median(np.abs(arr))
    # Or independently for each channel.
    else:
        median = np.median(np.abs(arr), axis=0)

    # Compute the threshold from the median.
    std = median / .6745
    threshold = std_factor * std
    assert isinstance(threshold, np.ndarray)

    if single_threshold:
        assert threshold.ndim == 1
        assert len(threshold) == 2
    else:
        assert threshold.ndim == 2
        assert threshold.shape == (2, nc)
    return threshold


class Thresholder(object):
    """Threshold traces to detect spikes.

    Parameters
    ----------

    mode : str
        `'positive'`, `'negative'`, or `'both'`.
    thresholds : dict
        A `{str: float}` mapping for multiple thresholds (e.g. `weak`
        and `strong`).

    Example
    -------

    ```python
    thres = Thresholder('positive', thresholds=(.1, .2))
    crossings = thres(traces)
    ```

    """
    def __init__(self,
                 mode=None,
                 thresholds=None,
                 ):
        assert mode in ('positive', 'negative', 'both')
        if isinstance(thresholds, (float, int, np.ndarray)):
            thresholds = {'default': thresholds}
        if thresholds is None:
            thresholds = {}
        assert isinstance(thresholds, dict)
        self._mode = mode
        self._thresholds = thresholds

    def transform(self, data):
        """Return `data`, `-data`, or `abs(data)` depending on the mode."""
        if self._mode == 'positive':
            return data
        elif self._mode == 'negative':
            return -data
        elif self._mode == 'both':
            return np.abs(data)

    def detect(self, data_t, threshold=None):
        """Perform the thresholding operation."""
        # Accept dictionary of thresholds.
        if isinstance(threshold, (list, tuple)):
            return {name: self(data_t, threshold=name)
                    for name in threshold}
        # Use the only threshold by default (if there is only one).
        if threshold is None:
            assert len(self._thresholds) == 1
            threshold = list(self._thresholds.keys())[0]
        # Fetch the threshold from its name.
        if isinstance(threshold, string_types):
            assert threshold in self._thresholds
            threshold = self._thresholds[threshold]
        # threshold = float(threshold)
        # Threshold the data.
        return data_t > threshold

    def __call__(self, data, threshold=None):
        # Transform the data according to the mode.
        data_t = self.transform(data)
        return self.detect(data_t, threshold=threshold)


# -----------------------------------------------------------------------------
# Connected components
# -----------------------------------------------------------------------------

def _to_tuples(x):
    return ((i, j) for (i, j) in x)


def _to_list(x):
    return [(i, j) for (i, j) in x]


def connected_components(weak_crossings=None,
                         strong_crossings=None,
                         probe_adjacency_list=None,
                         join_size=None,
                         channels=None):
    """Find all connected components in binary arrays of threshold crossings.

    Parameters
    ----------

    weak_crossings : array
        `(n_samples, n_channels)` array with weak threshold crossings
    strong_crossings : array
        `(n_samples, n_channels)` array with strong threshold crossings
    probe_adjacency_list : dict
        A dict `{channel: [neighbors]}`
    channels : array
        An (n_channels,) array with a list of all non-dead channels
    join_size : int
        The number of samples defining the tolerance in time for
        finding connected components


    Returns
    -------

    A list of lists of pairs `(samp, chan)` of the connected components in
    the 2D array `weak_crossings`, where a pair is adjacent if the samples are
    within `join_size` of each other, and the channels are adjacent in
    `probe_adjacency_list`, the channel graph.

    Note
    ----

    The channel mapping assumes that column #i in the data array is channel #i
    in the probe adjacency graph.

    """

    if probe_adjacency_list is None:
        probe_adjacency_list = {}

    if channels is None:
        channels = []

    # If the channels aren't referenced at all but exist in 'channels', add a
    # trivial self-connection so temporal floodfill will work. If this channel
    # is dead, it should be removed from 'channels'.
    probe_adjacency_list.update({i: {i} for i in channels
                                if not probe_adjacency_list.get(i)})

    # Make sure the values are sets.
    probe_adjacency_list = {c: set(cs)
                            for c, cs in probe_adjacency_list.items()}

    if strong_crossings is None:
        strong_crossings = weak_crossings

    assert weak_crossings.shape == strong_crossings.shape

    # Set of connected component labels which contain at least one strong
    # node.
    strong_nodes = set()

    n_s, n_ch = weak_crossings.shape
    join_size = int(join_size or 0)

    # An array with the component label for each node in the array
    label_buffer = np.zeros((n_s, n_ch), dtype=np.int32)

    # Component indices, a dictionary with keys the label of the component
    # and values a list of pairs (sample, channel) belonging to that component
    comp_inds = {}

    # mgraph is the channel graph, but with edge node connected to itself
    # because we want to include ourself in the adjacency. Each key of the
    # channel graph (a dictionary) is a node, and the value is a set of nodes
    # which are connected to it by an edge
    mgraph = {}
    for source, targets in probe_adjacency_list.items():
        # we add self connections
        mgraph[source] = targets.union([source])

    # Label of the next component
    c_label = 1

    # For all pairs sample, channel which are nonzero (note that numpy .nonzero
    # returns (all_i_s, all_i_ch), a pair of lists whose values at the
    # corresponding place are the sample, channel pair which is nonzero. The
    # lists are also returned in sorted order, so that i_s is always increasing
    # and i_ch is always increasing for a given value of i_s. izip is an
    # iterator version of the Python zip function, i.e. does the same as zip
    # but quicker. zip(A,B) is a list of all pairs (a,b) with a in A and b in B
    # in order (i.e. (A[0], B[0]), (A[1], B[1]), .... In conclusion, the next
    # line loops through all the samples i_s, and for each sample it loops
    # through all the channels.
    for i_s, i_ch in zip(*weak_crossings.nonzero()):
        # The next two lines iterate through all the neighbours of i_s, i_ch
        # in the graph defined by graph in the case of edges, and
        # j_s from i_s-join_size to i_s.
        for j_s in range(i_s - join_size, i_s + 1):
            # Allow us to leave out a channel from the graph to exclude bad
            # channels
            if i_ch not in mgraph:
                continue
            for j_ch in mgraph[i_ch]:
                # Label of the adjacent element.
                adjlabel = label_buffer[j_s, j_ch]
                # If the adjacent element is nonzero we need to do something.
                if adjlabel:
                    curlabel = label_buffer[i_s, i_ch]
                    if curlabel == 0:
                        # If current element is still zero, we just assign
                        # the label of the adjacent element to the current one.
                        label_buffer[i_s, i_ch] = adjlabel
                        # And add it to the list for the labelled component.
                        comp_inds[adjlabel].append((i_s, i_ch))

                    elif curlabel != adjlabel:
                        # If the current element is unequal to the adjacent
                        # one, we merge them by reassigning the elements of the
                        # adjacent component to the current one.
                        # samps_chans is an array of pairs sample, channel
                        # currently assigned to component adjlabel.
                        samps_chans = np.array(comp_inds[adjlabel],
                                               dtype=np.int32)

                        # samps_chans[:, 0] is the sample indices, so this
                        # gives only the samp,chan pairs that are within
                        # join_size of the current point.
                        # TODO: is this the right behaviour? If a component can
                        # have a width bigger than join_size I think it isn't!
                        samps_chans = samps_chans[i_s - samps_chans[:, 0] <=
                                                  join_size]

                        # Relabel the adjacent samp,chan points with current
                        # label.
                        samps, chans = samps_chans[:, 0], samps_chans[:, 1]
                        label_buffer[samps, chans] = curlabel

                        # Add them to the current label list, and remove the
                        # adjacent component entirely.
                        comp_inds[curlabel].extend(comp_inds.pop(adjlabel))

                        # Did not deal with merge condition, now fixed it
                        # seems...
                        # WARNING: might this "in" incur a performance hit
                        # here...?
                        if adjlabel in strong_nodes:
                            strong_nodes.add(curlabel)
                            strong_nodes.remove(adjlabel)

                    # NEW: add the current component label to the set of all
                    # strong nodes, if the current node is strong.
                    if curlabel > 0 and strong_crossings[i_s, i_ch]:
                        strong_nodes.add(curlabel)

        if label_buffer[i_s, i_ch] == 0:
            # If nothing is adjacent, we have the beginnings of a new
            # component, # so we label it, create a new list for the new
            # component which is given label c_label,
            # then increase c_label for the next new component afterwards.
            label_buffer[i_s, i_ch] = c_label
            comp_inds[c_label] = [(i_s, i_ch)]
            if strong_crossings[i_s, i_ch]:
                strong_nodes.add(c_label)
            c_label += 1

    # Only return the values, because we don't actually need the labels.
    comps = [comp_inds[key] for key in comp_inds.keys() if key in strong_nodes]
    return comps


class FloodFillDetector(object):
    """Detect spikes in weak and strong threshold crossings.

    Parameters
    ----------

    probe_adjacency_list : dict
        A dict `{channel: [neighbors]}`.
    join_size : int
        The number of samples defining the tolerance in time for
        finding connected components

    Example
    -------

    ```python
    det = FloodFillDetector(probe_adjacency_list=...,
                            join_size=...)
    components = det(weak_crossings, strong_crossings)
    ```

    `components` is a list of `(n, 2)` int arrays with the sample and channel
    for every sample in the component.

    """
    def __init__(self, probe_adjacency_list=None, join_size=None,
                 channels_per_group=None):
        self._adjacency_list = probe_adjacency_list
        self._join_size = join_size
        self._channels_per_group = channels_per_group

    def __call__(self, weak_crossings=None, strong_crossings=None):
        weak_crossings = np.asarray(weak_crossings, np.bool)
        strong_crossings = np.asarray(strong_crossings, np.bool)
        all_channels = sorted([item for sublist
                              in self._channels_per_group.values()
                              for item in sublist])

        cc = connected_components(weak_crossings=weak_crossings,
                                  strong_crossings=strong_crossings,
                                  probe_adjacency_list=self._adjacency_list,
                                  channels=all_channels,
                                  join_size=self._join_size,
                                  )
        # cc is a list of list of pairs (sample, channel)
        return [np.array(c) for c in cc]



# -*- coding: utf-8 -*-

"""Spike detection store."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from collections import defaultdict
import logging
import os
import os.path as op

import numpy as np
from six import string_types

from ..utils import (_save_arrays,
                     _load_arrays,
                     _concatenate,
                     _ensure_dir_exists,
                     )

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Spike counts
#------------------------------------------------------------------------------

class SpikeCounts(object):
    """Count spikes in chunks and channel groups."""
    def __init__(self, counts=None, groups=None, chunk_keys=None):
        self._groups = groups
        self._chunk_keys = chunk_keys
        self._counts = counts or defaultdict(lambda: defaultdict(int))

    def append(self, group=None, chunk_key=None, count=None):
        self._counts[group][chunk_key] += count

    @property
    def counts(self):
        return self._counts

    def per_group(self, group):
        return sum(self._counts.get(group, {}).values())

    def per_chunk(self, chunk_key):
        return sum(self._counts[group].get(chunk_key, 0)
                   for group in self._groups)

    def __call__(self, group=None, chunk_key=None):
        if group is not None and chunk_key is not None:
            return self._counts.get(group, {}).get(chunk_key, 0)
        elif group is not None:
            return self.per_group(group)
        elif chunk_key is not None:
            return self.per_chunk(chunk_key)
        elif group is None and chunk_key is None:
            return sum(self.per_group(group) for group in self._groups)


#------------------------------------------------------------------------------
# Spike detection store
#------------------------------------------------------------------------------

class ArrayStore(object):
    def __init__(self, root_dir):
        self._root_dir = op.realpath(root_dir)
        _ensure_dir_exists(self._root_dir)

    def _rel_path(self, **kwargs):
        """Relative to the root."""
        raise NotImplementedError()

    def _path(self, **kwargs):
        """Absolute path of a data file."""
        path = op.realpath(op.join(self._root_dir, self._rel_path(**kwargs)))
        _ensure_dir_exists(op.dirname(path))
        assert path.endswith('.npy')
        return path

    def _offsets_path(self, path):
        assert path.endswith('.npy')
        return op.splitext(path)[0] + '.offsets.npy'

    def _contains_multiple_arrays(self, path):
        return op.exists(path) and op.exists(self._offsets_path(path))

    def store(self, data=None, **kwargs):
        """Store an array or list of arrays."""
        path = self._path(**kwargs)
        if isinstance(data, list):
            if not data:
                return
            _save_arrays(path, data)
        elif isinstance(data, np.ndarray):
            dtype = data.dtype
            if not data.size:
                return
            assert dtype != np.object
            np.save(path, data)
        # debug("Store {}.".format(path))

    def load(self, **kwargs):
        path = self._path(**kwargs)
        if not op.exists(path):
            logger.log(5, "File `{}` doesn't exist.".format(path))
            return
        # Multiple arrays:
        # debug("Load {}.".format(path))
        if self._contains_multiple_arrays(path):
            return _load_arrays(path)
        else:
            return np.load(path)

    def delete(self, **kwargs):
        path = self._path(**kwargs)
        if op.exists(path):
            os.remove(path)
            # debug("Deleted `{}`.".format(path))
        offsets_path = self._offsets_path(path)
        if op.exists(offsets_path):
            os.remove(offsets_path)
            # debug("Deleted `{}`.".format(offsets_path))


class SpikeDetektStore(ArrayStore):
    """Store the following items:

    * filtered
    * components
    * spike_samples
    * features
    * masks

    """
    def __init__(self, root_dir, groups=None, chunk_keys=None):
        super(SpikeDetektStore, self).__init__(root_dir)
        self._groups = groups
        self._chunk_keys = chunk_keys
        self._spike_counts = SpikeCounts(groups=groups, chunk_keys=chunk_keys)

    def _rel_path(self, name=None, chunk_key=None, group=None):
        assert chunk_key >= 0
        assert group is None or group >= 0
        assert isinstance(name, string_types)
        group = group if group is not None else 'all'
        return 'group_{group}/{name}/chunk_{chunk:d}.npy'.format(
            chunk=chunk_key, name=name, group=group)

    @property
    def groups(self):
        return self._groups

    @property
    def chunk_keys(self):
        return self._chunk_keys

    def _iter(self, group=None, name=None):
        for chunk_key in self.chunk_keys:
            yield self.load(group=group, chunk_key=chunk_key, name=name)

    def spike_samples(self, group=None):
        if group is None:
            return {group: self.spike_samples(group) for group in self._groups}
        return self.concatenate(self._iter(group=group, name='spike_samples'))

    def features(self, group=None):
        """Yield chunk features."""
        if group is None:
            return {group: self.features(group) for group in self._groups}
        return self._iter(group=group, name='features')

    def masks(self, group=None):
        """Yield chunk masks."""
        if group is None:
            return {group: self.masks(group) for group in self._groups}
        return self._iter(group=group, name='masks')

    @property
    def spike_counts(self):
        return self._spike_counts

    def append(self, group=None, chunk_key=None,
               spike_samples=None, features=None, masks=None,
               spike_offset=0):
        if spike_samples is None or len(spike_samples) == 0:
            return
        n = len(spike_samples)
        assert features.shape[0] == n
        assert masks.shape[0] == n
        spike_samples = spike_samples + spike_offset
        assert np.all(np.diff(spike_samples) >= 0)

        self.store(group=group, chunk_key=chunk_key,
                   name='features', data=features)
        self.store(group=group, chunk_key=chunk_key,
                   name='masks', data=masks)
        self.store(group=group, chunk_key=chunk_key,
                   name='spike_samples', data=spike_samples)
        self._spike_counts.append(group=group, chunk_key=chunk_key, count=n)

    def concatenate(self, arrays):
        return _concatenate(arrays)

    def delete_all(self, name):
        """Delete all files for a given data name."""
        for group in self._groups:
            for chunk_key in self._chunk_keys:
                super(SpikeDetektStore, self).delete(name=name, group=group,
                                                     chunk_key=chunk_key)



# -*- coding: utf-8 -*-

"""Utility functions for test datasets."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import hashlib
import logging
import os
import os.path as op

from .utils import _klusta_user_dir, _ensure_dir_exists

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Utility functions
#------------------------------------------------------------------------------

_BASE_URL = {
    'cortexlab': 'http://phy.cortexlab.net/data/samples/',
    'github': 'https://raw.githubusercontent.com/kwikteam/phy-data/master/',
    'local': 'http://localhost:8000/',
}


def _remote_file_size(path):
    import requests
    try:
        response = requests.head(path)
        return int(response.headers.get('content-length', 0))
    except Exception:
        # Unable to get the file size: no progress report.
        pass
    return 0


def _save_stream(r, path):
    size = _remote_file_size(r.url)
    # pr = ProgressReporter()
    # pr.value_max = size or 1
    # pr.set_progress_message('Downloading `' + path + '`: {progress:.1f}%.')
    # pr.set_complete_message('Download complete.')
    downloaded = 0
    with open(path, 'wb') as f:
        for i, chunk in enumerate(r.iter_content(chunk_size=1024)):
            if chunk:
                f.write(chunk)
                f.flush()
                downloaded += len(chunk)
                # if i % 100 == 0:
                #     pr.value = downloaded
    if size:
        assert size == downloaded
    # pr.set_complete()


def _download(url, stream=None):
    from requests import get
    r = get(url, stream=stream)
    if r.status_code != 200:
        logger.debug("Error while downloading `%s`.", url)
        r.raise_for_status()
    return r


def download_text_file(url):
    """Download a text file."""
    return _download(url).text


def _md5(path, blocksize=2 ** 20):
    """Compute the checksum of a file."""
    m = hashlib.md5()
    with open(path, 'rb') as f:
        while True:
            buf = f.read(blocksize)
            if not buf:
                break
            m.update(buf)
    return m.hexdigest()


def _check_md5(path, checksum):
    if checksum is None:
        return
    return _md5(path) == checksum


def _check_md5_of_url(output_path, url):
    try:
        checksum = download_text_file(url + '.md5').split(' ')[0]
    except Exception:
        checksum = None
    finally:
        if checksum:
            return _check_md5(output_path, checksum)


def _validate_output_dir(output_dir):
    if output_dir is None:
        output_dir = '.'
    if not output_dir.endswith('/'):
        output_dir = output_dir + '/'
    output_dir = op.realpath(op.dirname(output_dir))
    if not op.exists(output_dir):
        os.mkdir(output_dir)
    return output_dir


def download_file(url, output_path=None):
    """Download a binary file from an URL.

    The checksum will be downloaded from `URL + .md5`. If this download
    succeeds, the file's MD5 will be compared to the expected checksum.

    Parameters
    ----------

    url : str
        The file's URL.
    output_path : str or None
        The path where the file is to be saved.

    Returns
    -------

    output_path : str
        The path where the file was downloaded.

    """
    if output_path is None:
        output_path = url.split('/')[-1]
    if op.exists(output_path):
        checked = _check_md5_of_url(output_path, url)
        if checked is False:
            logger.debug("The file `%s` already exists but is invalid:"
                         " redownloading.", output_path)
        elif checked is True:
            logger.debug("The file `%s` already exists: skipping.",
                         output_path)
            return
    r = _download(url, stream=True)
    _save_stream(r, output_path)
    if _check_md5_of_url(output_path, url) is False:
        logger.debug("The checksum doesn't match: retrying the download.")
        r = _download(url, stream=True)
        _save_stream(r, output_path)
        if _check_md5_of_url(output_path, url) is False:
            raise RuntimeError("The checksum of the downloaded file "
                               "doesn't match the provided checksum.")
    return output_path


def download_test_data(name, phy_user_dir=None, force=False):
    """Download a test file."""
    phy_user_dir = phy_user_dir or _klusta_user_dir()
    dir = op.join(phy_user_dir, 'test_data')
    _ensure_dir_exists(dir)
    path = op.join(dir, name)
    if not force and op.exists(path):
        return path
    url = _BASE_URL['github'] + 'test/' + name
    download_file(url, output_path=path)
    return path


def download_sample_data(filename, output_dir=None, base='cortexlab'):
    """Download a sample dataset.

    Parameters
    ----------

    filename : str
        Name of the sample dataset to download.
    output_dir : str
        The directory where to save the file.
    base : str
        The id of the base URL. Can be `'cortexlab'` or `'github'`.

    """
    output_dir = _validate_output_dir(output_dir)
    url = _BASE_URL[base] + filename
    output_path = op.join(output_dir, filename)
    try:
        download_file(url, output_path=output_path)
    except Exception as e:
        logger.warn("An error occurred while downloading `%s` to `%s`: %s",
                    url, output_path, str(e))



# -*- coding: utf-8 -*-

"""Tests of HDF5 routines."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os.path as op

import numpy as np
from numpy.testing import assert_array_equal as ae
from pytest import raises

from ...utils import captured_output
from ..h5 import open_h5, _split_hdf5_path


#------------------------------------------------------------------------------
# Utility test routines
#------------------------------------------------------------------------------

def _create_test_file(dirpath):
    filename = op.join(dirpath, '_test.h5')
    with open_h5(filename, 'w') as tempfile:
        # Create a random dataset using h5py directly.
        h5file = tempfile.h5py_file
        h5file.create_dataset('ds1', (10,), dtype=np.float32)
        group = h5file.create_group('/mygroup')
        h5file.create_dataset('/mygroup/ds2', (10,), dtype=np.int8)
        group.attrs['myattr'] = 123
        return tempfile.filename


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_split_hdf5_path():
    # The path should always start with a leading '/'.
    with raises(ValueError):
        _split_hdf5_path('')
    with raises(ValueError):
        _split_hdf5_path('path')

    h, t = _split_hdf5_path('/')
    assert (h == '/') and (t == '')

    h, t = _split_hdf5_path('/path')
    assert (h == '/') and (t == 'path')

    h, t = _split_hdf5_path('/path/')
    assert (h == '/path') and (t == '')

    h, t = _split_hdf5_path('/path/to')
    assert (h == '/path') and (t == 'to')

    h, t = _split_hdf5_path('/path/to/')
    assert (h == '/path/to') and (t == '')

    # Check that invalid paths raise errors.
    with raises(ValueError):
        _split_hdf5_path('path/')
    with raises(ValueError):
        _split_hdf5_path('/path//')
    with raises(ValueError):
        _split_hdf5_path('/path//to')


def test_h5_read(tempdir):
    # Create the test HDF5 file in the temporary directory.
    filename = _create_test_file(tempdir)

    # Test close() method.
    f = open_h5(filename)
    assert f.is_open()
    f.close()
    assert not f.is_open()
    with raises(IOError):
        f.describe()

    # Open the test HDF5 file.
    with open_h5(filename) as f:
        assert f.is_open()

        assert f.children() == ['ds1', 'mygroup']
        assert f.groups() == ['mygroup']
        assert f.datasets() == ['ds1']
        assert f.attrs('/mygroup') == ['myattr']
        assert f.attrs('/mygroup_nonexisting') == []

        # Check dataset ds1.
        ds1 = f.read('/ds1')[:]
        assert isinstance(ds1, np.ndarray)
        assert ds1.shape == (10,)
        assert ds1.dtype == np.float32

        # Check dataset ds2.
        ds2 = f.read('/mygroup/ds2')[:]
        assert isinstance(ds2, np.ndarray)
        assert ds2.shape == (10,)
        assert ds2.dtype == np.int8

        # Check HDF5 group attribute.
        assert f.has_attr('/mygroup', 'myattr')
        assert not f.has_attr('/mygroup', 'myattr_bis')
        assert not f.has_attr('/mygroup_bis', 'myattr_bis')
        value = f.read_attr('/mygroup', 'myattr')
        assert value == 123

        # Check that errors are raised when the paths are invalid.
        with raises(Exception):
            f.read('//path')
        with raises(Exception):
            f.read('/path//')
        with raises(ValueError):
            f.read('/nonexistinggroup')
        with raises(ValueError):
            f.read('/nonexistinggroup/ds34')

    assert not f.is_open()


def test_h5_append(tempdir):
    # Create the test HDF5 file in the temporary directory.
    filename = _create_test_file(tempdir)

    with open_h5(filename, 'a') as f:
        f.write('/ds_empty', dtype=np.float32, shape=(10, 2))
        arr = f.read('/ds_empty')
        arr[:5, 0] = 1

    with open_h5(filename, 'r') as f:
        arr = f.read('/ds_empty')[...]
        assert np.all(arr[:5, 0] == 1)


def test_h5_write(tempdir):
    # Create the test HDF5 file in the temporary directory.
    filename = _create_test_file(tempdir)

    # Create some array.
    temp_array = np.zeros(10, dtype=np.float32)

    # Open the test HDF5 file in read-only mode (the default) and
    # try to write in it. This should raise an exception.
    with open_h5(filename) as f:
        with raises(Exception):
            f.write('/ds1', temp_array)

    # Open the test HDF5 file in read/write mode and
    # try to write in an existing dataset.
    with open_h5(filename, 'a') as f:
        # This raises an exception because the file already exists,
        # and by default this is forbidden.
        with raises(ValueError):
            f.write('/ds1', temp_array)

        # This works, though, because we force overwriting the dataset.
        f.write('/ds1', temp_array, overwrite=True)
        ae(f.read('/ds1'), temp_array)

        # Write a new array.
        f.write('/ds2', temp_array)
        ae(f.read('/ds2'), temp_array)

        # Write a new array in a nonexistent group.
        f.write('/ds3/ds4/ds5', temp_array)
        ae(f.read('/ds3/ds4/ds5'), temp_array)

        # Write an existing attribute.
        f.write_attr('/ds1', 'myattr', 456)

        with raises(KeyError):
            f.read_attr('/ds1', 'nonexistingattr')

        assert f.read_attr('/ds1', 'myattr') == 456

        # Write a new attribute in a dataset.
        f.write_attr('/ds1', 'mynewattr', 789)
        assert f.read_attr('/ds1', 'mynewattr') == 789

        # Write a new attribute in a group.
        f.write_attr('/mygroup', 'mynewattr', 890)
        assert f.read_attr('/mygroup', 'mynewattr') == 890

        # Write a new attribute in a nonexisting group.
        f.write_attr('/nonexistinggroup', 'mynewattr', 2)
        assert f.read_attr('/nonexistinggroup', 'mynewattr') == 2

        # Write a new attribute two levels into a nonexisting group.
        f.write_attr('/nonexistinggroup2/group3', 'mynewattr', 2)
        assert f.read_attr('/nonexistinggroup2/group3', 'mynewattr') == 2


def test_h5_describe(tempdir):
    # Create the test HDF5 file in the temporary directory.
    filename = _create_test_file(tempdir)

    # Open the test HDF5 file.
    with open_h5(filename) as f:
        with captured_output() as (out, err):
            f.describe()
    output = out.getvalue().strip()
    output_lines = output.split('\n')
    assert len(output_lines) == 3


def test_h5_move(tempdir):
    # Create the test HDF5 file in the temporary directory.
    filename = _create_test_file(tempdir)

    with open_h5(filename, 'a') as f:

        # Test dataset move.
        assert f.exists('ds1')
        arr = f.read('ds1')[:]
        assert len(arr) == 10
        f.move('ds1', 'ds1_new')
        assert not f.exists('ds1')
        assert f.exists('ds1_new')
        arr_new = f.read('ds1_new')[:]
        assert len(arr_new) == 10
        ae(arr, arr_new)

        # Test group move.
        assert f.exists('mygroup/ds2')
        arr = f.read('mygroup/ds2')
        f.move('mygroup', 'g/mynewgroup')
        assert not f.exists('mygroup')
        assert f.exists('g/mynewgroup')
        assert f.exists('g/mynewgroup/ds2')
        arr_new = f.read('g/mynewgroup/ds2')
        ae(arr, arr_new)


def test_h5_copy(tempdir):
    # Create the test HDF5 file in the temporary directory.
    filename = _create_test_file(tempdir)

    with open_h5(filename, 'a') as f:

        # Test dataset copy.
        assert f.exists('ds1')
        arr = f.read('ds1')[:]
        assert len(arr) == 10
        f.copy('ds1', 'ds1_new')
        assert f.exists('ds1')
        assert f.exists('ds1_new')
        arr_new = f.read('ds1_new')[:]
        assert len(arr_new) == 10
        ae(arr, arr_new)

        # Test group copy.
        assert f.exists('mygroup/ds2')
        arr = f.read('mygroup/ds2')
        f.copy('mygroup', 'g/mynewgroup')
        assert f.exists('mygroup')
        assert f.exists('g/mynewgroup')
        assert f.exists('g/mynewgroup/ds2')
        arr_new = f.read('g/mynewgroup/ds2')
        ae(arr, arr_new)


def test_h5_delete(tempdir):
    # Create the test HDF5 file in the temporary directory.
    filename = _create_test_file(tempdir)

    with open_h5(filename, 'a') as f:

        # Test dataset delete.
        assert f.exists('ds1')
        with raises(ValueError):
            f.delete('a')
        f.delete('ds1')
        assert not f.exists('ds1')

        # Test group delete.
        assert f.exists('mygroup/ds2')
        f.delete('mygroup')
        assert not f.exists('mygroup')
        assert not f.exists('mygroup/ds2')



# -*- coding: utf-8 -*-

"""Default settings for spike detection."""


# -----------------------------------------------------------------------------
# Spike detection
# -----------------------------------------------------------------------------

spikedetekt = {
    'filter_low': 500.,
    'filter_high_factor': 0.95 * .5,  # will be multiplied by the sample rate
    'filter_butter_order': 3,

    # Data chunks.
    'chunk_size_seconds': 1.,
    'chunk_overlap_seconds': .015,

    # Threshold.
    'n_excerpts': 50,
    'excerpt_size_seconds': 1.,
    'use_single_threshold': True,
    'threshold_strong_std_factor': 4.5,
    'threshold_weak_std_factor': 2.,
    'detect_spikes': 'negative',

    # Connected components.
    'connected_component_join_size': 1,

    # Spike extractions.
    'extract_s_before': 10,
    'extract_s_after': 10,
    'weight_power': 2,

    # Features.
    'n_features_per_channel': 3,
    'pca_n_waveforms_max': 10000,

}



# -*- coding: utf-8 -*-

"""Waveform filtering routines."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from scipy import signal


#------------------------------------------------------------------------------
# Waveform filtering routines
#------------------------------------------------------------------------------

def bandpass_filter(rate=None, low=None, high=None, order=None):
    """Butterworth bandpass filter."""
    assert low < high
    assert order >= 1
    return signal.butter(order,
                         (low / (rate / 2.), high / (rate / 2.)),
                         'pass')


def apply_filter(x, filter=None, axis=0):
    """Apply a filter to an array."""
    if isinstance(x, list):
        x = np.asarray(x)
    if x.shape[axis] == 0:
        return x
    b, a = filter
    return signal.filtfilt(b, a, x[:], axis=axis)


class Filter(object):
    """Multichannel bandpass filter.

    The filter is applied on every column of a 2D array.

    Example
    -------

    ```python
    fil = Filter(rate=20000., low=5000., high=15000., order=4)
    traces_f = fil(traces)
    ```

    """
    def __init__(self, rate=None, low=None, high=None, order=None):
        self._filter = bandpass_filter(rate=rate,
                                       low=low,
                                       high=high,
                                       order=order,
                                       )

    def __call__(self, data):
        return apply_filter(data, filter=self._filter)


#------------------------------------------------------------------------------
# Whitening
#------------------------------------------------------------------------------

class Whitening(object):
    """Compute a whitening matrix and apply it to data.

    Contributed by Pierre Yger.

    """
    def fit(self, x, fudge=1e-18):
        """Compute the whitening matrix.

        Parameters
        ----------

        x : array
            An `(n_samples, n_channels)` array.

        """
        assert x.ndim == 2
        ns, nc = x.shape
        x_cov = np.cov(x, rowvar=0)
        assert x_cov.shape == (nc, nc)
        d, v = np.linalg.eigh(x_cov)
        d = np.diag(1. / np.sqrt(d + fudge))
        # This is equivalent, but seems much slower...
        # w = np.einsum('il,lk,jk->ij', v, d, v)
        w = np.dot(np.dot(v, d), v.T)
        self._matrix = w
        return w

    def transform(self, x):
        """Whiten some data.

        Parameters
        ----------

        x : array
            An `(n_samples, n_channels)` array.

        """
        return np.dot(x, self._matrix)



# -*- coding: utf-8 -*-

"""Tests of clustering algorithms."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_equal as ae
from pytest import mark

from ..spikedetekt import (SpikeDetekt,
                           _split_spikes,
                           _concat,
                           _concatenate,
                           _subtract_offsets,
                           _relative_channels,
                           )


#------------------------------------------------------------------------------
# Tests spike detection
#------------------------------------------------------------------------------

def test_subtract_offsets():
    samples = [2, 3, 5, 7, 11]

    offsets = [0, 20]
    s, r = _subtract_offsets(samples, offsets)
    ae(s, samples)
    ae(r, np.zeros(len(samples)))

    offsets = [0, 5, 10, 15, 20]
    s, r = _subtract_offsets(samples, offsets)
    ae(s, [2, 3, 0, 2, 1])
    ae(r, [0, 0, 1, 1, 2])


def test_relative_channels():
    ch = {0: [2, 3, 7, 5]}
    adj = {2: set([3]), 3: set([2, 7]), 5: set([5]), 7: set([])}
    ch, adj = _relative_channels(ch, adj)
    assert ch == {0: [0, 1, 2, 3]}
    assert adj == {0: set([1]), 1: set([0, 2]), 2: set([]), 3: set([3])}


def test_split_spikes():
    groups = np.zeros(10, dtype=np.int)
    groups[1::2] = 1

    idx = np.ones(10, dtype=np.bool)
    idx[0] = False
    idx[-1] = False

    a = np.random.rand(10, 2)
    b = np.random.rand(10, 3, 2)

    out = _split_spikes(groups, idx, a=a, b=b)

    assert sorted(out) == [0, 1]
    assert sorted(out[0]) == ['a', 'b']
    assert sorted(out[1]) == ['a', 'b']

    ae(out[0]['a'], a[1:-1][1::2])
    ae(out[0]['b'], b[1:-1][1::2])

    ae(out[1]['a'], a[1:-1][::2])
    ae(out[1]['b'], b[1:-1][::2])


def test_spike_detect_methods(tempdir, raw_dataset):
    params = raw_dataset.params
    probe = raw_dataset.probe
    sample_rate = raw_dataset.sample_rate
    sd = SpikeDetekt(tempdir=tempdir,
                     probe=raw_dataset.probe,
                     sample_rate=sample_rate,
                     **params)
    traces = raw_dataset.traces
    n_samples = raw_dataset.n_samples
    n_channels = raw_dataset.n_channels

    # Filter the data.
    traces_f = sd.apply_filter(traces)
    assert traces_f.shape == traces.shape
    assert not np.any(np.isnan(traces_f))

    # Thresholds.
    thresholds = sd.find_thresholds(traces)
    assert np.all(0 <= thresholds['weak'])
    assert np.all(thresholds['weak'] <= thresholds['strong'])

    # Spike detection.
    traces_f[1000:1010, :3] *= 5
    traces_f[2000:2010, [0, 2]] *= 5
    traces_f[3000:3020, :] *= 5
    components = sd.detect(traces_f, thresholds)
    assert isinstance(components, list)
    # n_spikes = len(components)
    n_samples_waveforms = (params['extract_s_before'] +
                           params['extract_s_after'])

    # Spike extraction.
    split = sd.extract_spikes(components, traces_f, thresholds,
                              keep_bounds=(0, n_samples))

    if not split:
        return
    samples = _concat(split[0]['spike_samples'], np.float64)
    waveforms = _concat(split[0]['waveforms'], np.float32)
    masks = _concat(split[0]['masks'], np.float32)

    n_spikes = len(samples)
    n_channels = len(probe['channel_groups'][0]['channels'])

    assert samples.dtype == np.float64
    assert samples.shape == (n_spikes,)
    assert waveforms.shape == (n_spikes, n_samples_waveforms, n_channels)
    assert masks.shape == (n_spikes, n_channels)
    assert 0. <= masks.min() < masks.max() <= 1.
    assert not np.any(np.isnan(samples))
    assert not np.any(np.isnan(waveforms))
    assert not np.any(np.isnan(masks))

    # PCA.
    pcs = sd.waveform_pcs(waveforms, masks)
    n_pcs = params['n_features_per_channel']
    assert pcs.shape == (n_pcs, n_samples_waveforms, n_channels)
    assert not np.any(np.isnan(pcs))

    # Features.
    features = sd.features(waveforms, pcs)
    assert features.shape == (n_spikes, n_channels, n_pcs)
    assert not np.any(np.isnan(features))


@mark.long
def test_spike_detect_real_data(tempdir, raw_dataset):

    params = raw_dataset.params
    probe = raw_dataset.probe
    sample_rate = raw_dataset.sample_rate
    sd = SpikeDetekt(tempdir=tempdir,
                     probe=probe,
                     sample_rate=sample_rate,
                     **params)
    traces = raw_dataset.traces
    npc = params['n_features_per_channel']
    # n_samples_w = params['extract_s_before'] + params['extract_s_after']

    # Run the detection.
    out = sd.run_serial(traces, interval_samples=(0, 60000))

    channels = probe['channel_groups'][0]['channels']
    n_channels = len(channels)

    spike_samples = _concatenate(out.spike_samples[0])
    masks = _concatenate(out.masks[0])
    features = _concatenate(out.features[0])
    n_spikes = out.n_spikes_per_group[0]

    if n_spikes:
        assert spike_samples.shape == (n_spikes,)
        assert masks.shape == (n_spikes, n_channels)
        assert features.shape == (n_spikes, n_channels, npc)

        # There should not be any spike with only masked channels.
        assert np.all(masks.max(axis=1) > 0)

    # # Plot...
    # from phy.plot.traces import plot_traces
    # c = plot_traces(traces[:30000, channels],
    #                 spike_samples=spike_samples,
    #                 masks=masks,
    #                 n_samples_per_spike=n_samples_w,
    #                 show=False)



# -*- coding: utf-8 -*-

"""Tests of spike detection store."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from pytest import yield_fixture
import numpy as np
from numpy.testing import assert_equal as ae

from ..store import SpikeCounts, ArrayStore, SpikeDetektStore


#------------------------------------------------------------------------------
# Tests spike detection store
#------------------------------------------------------------------------------

@yield_fixture(params=['from_dict', 'append'])
def spike_counts(request):
    groups = [0, 2]
    chunk_keys = [10, 20, 30]
    if request.param == 'from_dict':
        c = {0: {10: 100, 20: 200},
             2: {10: 1, 30: 300},
             }
        sc = SpikeCounts(c, groups=groups, chunk_keys=chunk_keys)
    elif request.param == 'append':
        sc = SpikeCounts(groups=groups, chunk_keys=chunk_keys)
        sc.append(group=0, chunk_key=10, count=100)
        sc.append(group=0, chunk_key=20, count=200)
        sc.append(group=2, chunk_key=10, count=1)
        sc.append(group=2, chunk_key=30, count=300)
    yield sc


def test_spike_counts(spike_counts):
    assert spike_counts() == 601

    assert spike_counts(group=0) == 300
    assert spike_counts(group=1) == 0
    assert spike_counts(group=2) == 301

    assert spike_counts(chunk_key=10) == 101
    assert spike_counts(chunk_key=20) == 200
    assert spike_counts(chunk_key=30) == 300


class _TestArrayStore(ArrayStore):
    def _rel_path(self, a=None, b=None):
        return '{}/{}.npy'.format(a or 'none_a', b or 'none_b')


def test_array_store(tempdir):
    store = _TestArrayStore(tempdir)

    store.store(a=1, b=1, data=np.arange(11))
    store.store(a=1, b=2, data=np.arange(12))
    store.store(b=2, data=np.arange(2))
    store.store(b=3, data=None)
    store.store(a=1, data=[np.arange(3), np.arange(3, 8)])

    ae(store.load(a=0, b=1), None)
    ae(store.load(a=1, b=1), np.arange(11))
    ae(store.load(a=1, b=2), np.arange(12))
    ae(store.load(b=2), np.arange(2))
    ae(store.load(b=3), None)
    ae(store.load(a=1)[0], np.arange(3))
    ae(store.load(a=1)[1], np.arange(3, 8))

    store.delete(b=2)
    store.delete(a=1, b=2)


def test_spikedetekt_store(tempdir):
    groups = [0, 2]
    chunk_keys = [10, 20, 30]

    n_channels = 4
    npc = 2

    _keys = [(0, 10), (0, 20), (2, 10), (2, 30)]
    _counts = [100, 200, 1, 300]

    # Generate random spike samples, features, and masks.
    s = {k: np.arange(c) for k, c in zip(_keys, _counts)}
    f = {k: np.random.rand(c, n_channels, npc)
         for k, c in zip(_keys, _counts)}
    m = {k: np.random.rand(c, n_channels)
         for k, c in zip(_keys, _counts)}

    store = SpikeDetektStore(tempdir, groups=groups, chunk_keys=chunk_keys)

    # Save data.
    for group, chunk_key in _keys:
        spike_samples = s.get((group, chunk_key), None)
        features = f.get((group, chunk_key), None)
        masks = m.get((group, chunk_key), None)

        store.append(group=group, chunk_key=chunk_key,
                     spike_samples=spike_samples,
                     features=features,
                     masks=masks,
                     spike_offset=chunk_key,
                     )

    # Load data.
    for group in groups:
        # Check spike samples.
        ae(store.spike_samples(group),
           np.hstack([s[key] + key[1] for key in _keys if key[0] == group]))

        # Check features and masks.
        for name in ('features', 'masks'):
            # Actual data.
            data_dict = f if name == 'features' else m
            # Stored data (generator).
            data_gen = getattr(store, name)(group)
            # Go through all chunks.
            for chunk_key, data in zip(chunk_keys, data_gen):
                if (group, chunk_key) in _keys:
                    ae(data_dict[group, chunk_key], data)
                else:
                    assert data is None

    # Test spike counts.
    test_spike_counts(store.spike_counts)



# -*- coding: utf-8 -*-

"""PCA tests."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np

from ...kwik.mock import artificial_waveforms, artificial_masks
from ..pca import PCA, _compute_pcs


#------------------------------------------------------------------------------
# Test PCA
#------------------------------------------------------------------------------

def test_pca():
    n_spikes = 100
    n_samples = 40
    n_channels = 12
    waveforms = artificial_waveforms(n_spikes, n_samples, n_channels)
    masks = artificial_masks(n_spikes, n_channels)

    pca = PCA(n_pcs=3)
    pcs = pca.fit(waveforms, masks)
    assert pcs.shape == (3, n_samples, n_channels)
    fet = pca.transform(waveforms)
    assert fet.shape == (n_spikes, n_channels, 3)


def test_compute_pcs():
    """Test PCA on a 2D array."""
    # Horizontal ellipsoid.
    x = np.random.randn(20000, 2) * np.array([[10., 1.]])
    # Rotate the points by pi/4.
    a = 1. / np.sqrt(2.)
    rot = np.array([[a, -a], [a, a]])
    x = np.dot(x, rot)
    # Compute the PCs.
    pcs = _compute_pcs(x[..., None])
    assert pcs.ndim == 3
    assert (np.abs(pcs) - a).max() < 1e-2


def test_compute_pcs_3d():
    """Test PCA on a 3D array."""
    x1 = np.random.randn(20000, 2) * np.array([[10., 1.]])
    x2 = np.random.randn(20000, 2) * np.array([[1., 10.]])
    x = np.dstack((x1, x2))
    # Compute the PCs.
    pcs = _compute_pcs(x)
    assert pcs.ndim == 3
    assert np.linalg.norm(pcs[0, :, 0] - np.array([-1., 0.])) < 1e-2
    assert np.linalg.norm(pcs[1, :, 0] - np.array([0., -1.])) < 1e-2
    assert np.linalg.norm(pcs[0, :, 1] - np.array([0, 1.])) < 1e-2
    assert np.linalg.norm(pcs[1, :, 1] - np.array([-1., 0.])) < 1e-2



# -*- coding: utf-8 -*-

"""Tests of clustering algorithms."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae
from numpy.testing import assert_array_almost_equal as aae

from ..kwik.model import KwikModel
from ..kwik.mock import create_mock_kwik
from ..klustakwik import klustakwik, sparsify_features_masks


#------------------------------------------------------------------------------
# Fixtures
#------------------------------------------------------------------------------

sample_rate = 10000
n_samples = 25000
n_channels = 4


#------------------------------------------------------------------------------
# Tests clustering
#------------------------------------------------------------------------------

def test_sparsify_features_masks():
    # Data as arrays
    fet = np.array([[1, 3, 5, 7, 11],
                    [6, 7, 8, 9, 10],
                    [11, 12, 13, 14, 15],
                    [16, 17, 18, 19, 20]], dtype=float)

    fmask = np.array([[1, 0.5, 0, 0, 0],
                      [0, 1, 1, 0, 0],
                      [0, 1, 1, 0, 0],
                      [0, 0, 0, 1, 1]], dtype=float)

    # Normalisation to [0, 1]
    fet_original = fet.copy()
    fet = (fet - np.amin(fet, axis=0)) / \
        (np.amax(fet, axis=0) - np.amin(fet, axis=0))

    nanmasked_fet = fet.copy()
    nanmasked_fet[fmask > 0] = np.nan

    # Correct computation of the corrected data and correction term
    x = fet
    w = fmask
    nu = np.nanmean(nanmasked_fet, axis=0)[np.newaxis, :]
    sigma2 = np.nanvar(nanmasked_fet, axis=0)[np.newaxis, :]
    y = w * x + (1 - w) * nu
    z = w * x * x + (1 - w) * (nu * nu + sigma2)
    correction_terms = z - y * y
    features = y

    data = sparsify_features_masks(fet_original, fmask)

    aae(data.noise_mean, np.nanmean(nanmasked_fet, axis=0))
    aae(data.noise_variance, np.nanvar(nanmasked_fet, axis=0))
    assert np.amin(data.features) == 0
    assert np.amax(data.features) == 1
    assert len(data.offsets) == 5

    for i in range(4):
        data_u = data.unmasked[data.offsets[i]:data.offsets[i + 1]]
        true_u, = fmask[i, :].nonzero()
        ae(data_u, true_u)
        data_f = data.features[data.offsets[i]:data.offsets[i + 1]]
        true_f = fet[i, data_u]
        ae(data_f, true_f)
        data_m = data.masks[data.offsets[i]:data.offsets[i + 1]]
        true_m = fmask[i, data_u]
        ae(data_m, true_m)

    # PART 2: Check that converting to SparseData is correct
    # compute unique masks and apply correction terms to data
    data = data.to_sparse_data()

    assert data.num_spikes == 4
    assert data.num_features == 5
    assert data.num_masks == 3

    for i in range(4):
        data_u = data.unmasked[data.unmasked_start[i]:data.unmasked_end[i]]
        true_u, = fmask[i, :].nonzero()
        ae(data_u, true_u)
        data_f = data.features[data.values_start[i]:data.values_end[i]]
        true_f = features[i, data_u]
        aae(data_f, true_f)
        data_c = data.correction_terms[data.values_start[i]:data.values_end[i]]
        true_c = correction_terms[i, data_u]
        aae(data_c, true_c)
        data_m = data.masks[data.values_start[i]:data.values_end[i]]
        true_m = fmask[i, data_u]
        aae(data_m, true_m)


def test_klustakwik(tempdir):
    n_spikes = 100
    filename = create_mock_kwik(tempdir,
                                n_clusters=1,
                                n_spikes=n_spikes,
                                n_channels=8,
                                n_features_per_channel=3,
                                n_samples_traces=5000)
    model = KwikModel(filename)

    spike_clusters, params = klustakwik(model, num_starting_clusters=10)
    assert params['num_starting_clusters'] == 10
    assert len(spike_clusters) == n_spikes

    spike_clusters, params = klustakwik(model, num_starting_clusters=10,
                                        spike_ids=range(100))
    assert len(spike_clusters) == 100



# -*- coding: utf-8 -*-

"""Temporary directory used in unit tests."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import warnings as _warnings
import os as _os
import sys as _sys
from tempfile import mkdtemp

import six


#------------------------------------------------------------------------------
# Temporary directory
#------------------------------------------------------------------------------

class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tempdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.

    The code comes from http://stackoverflow.com/a/19299884/1595060

    """
    def __init__(self, suffix="", prefix="tmp", dir=None):
        # The tmp dir can be specified in the PHY_TMP_DIR env variable.
        dir = dir or _os.environ.get('PHY_TMP_DIR', None)
        self._closed = False
        self.name = None  # Handle mkdtemp raising an exception
        self.name = mkdtemp(suffix, prefix, dir)

    def __enter__(self):
        return self.name

    def cleanup(self, _warn=False):
        if self.name and not self._closed:
            try:
                self._rmtree(self.name)
            except (TypeError, AttributeError) as ex:
                # Issue #10188: Emit a warning on stderr
                # if the directory could not be cleaned
                # up due to missing globals
                if "None" not in str(ex):
                    raise
                six.print_("ERROR: {!r} while cleaning up {!r}".format(ex,
                                                                       self,),
                           file=_sys.stderr)
                return
            self._closed = True
            if _warn:
                # This should be a ResourceWarning, but it is not available in
                # Python 2.x.
                self._warn("Implicitly cleaning up {!r}".format(self),
                           Warning)

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def __del__(self):
        # Issue a ResourceWarning if implicit cleanup needed
        self.cleanup(_warn=True)

    # XXX (ncoghlan): The following code attempts to make
    # this class tolerant of the module nulling out process
    # that happens during CPython interpreter shutdown
    # Alas, it doesn't actually manage it. See issue #10188
    _listdir = staticmethod(_os.listdir)
    _path_join = staticmethod(_os.path.join)
    _isdir = staticmethod(_os.path.isdir)
    _islink = staticmethod(_os.path.islink)
    _remove = staticmethod(_os.remove)
    _rmdir = staticmethod(_os.rmdir)
    _warn = _warnings.warn

    def _rmtree(self, path):
        # Essentially a stripped down version of shutil.rmtree.  We can't
        # use globals because they may be None'ed out at shutdown.
        for name in self._listdir(path):
            fullname = self._path_join(path, name)
            try:
                isdir = self._isdir(fullname) and not self._islink(fullname)
            except OSError:
                isdir = False
            if isdir:
                self._rmtree(fullname)
            else:
                try:
                    self._remove(fullname)
                except OSError:
                    pass
        try:
            self._rmdir(path)
        except OSError:
            pass



# -*- coding: utf-8 -*-

"""Kwik creator."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os
import os.path as op
import logging

import numpy as np
from h5py import Dataset
from six import string_types
from six.moves import zip

from .h5 import open_h5
from .mea import load_probe
from .model import _DEFAULT_GROUPS
from ..utils import (_read_python,
                     _unique,
                     _dat_n_samples,
                     _ensure_dir_exists,
                     )
from .. import __version_git__

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Kwik creator
#------------------------------------------------------------------------------

def _write_by_chunk(dset, arrs):
    # Note: arrs should be a generator for performance reasons.
    assert isinstance(dset, Dataset)
    # Start the data.
    offset = 0
    for arr in arrs:
        n = arr.shape[0]
        arr = arr[...]
        # Match the shape of the chunk array with the dset shape.
        assert arr.shape == (n,) + dset.shape[1:]
        dset[offset:offset + n, ...] = arr
        offset += arr.shape[0]
    # Check that the copy is complete.
    assert offset == dset.shape[0]


def _concat(arrs):
    return np.hstack([_[...] for _ in arrs])


class KwikCreator(object):
    """Create and modify a `.kwik` file."""
    def __init__(self, basename=None, kwik_path=None, kwx_path=None):
        # Find the .kwik filename.
        if kwik_path is None:
            assert basename is not None
            if basename.endswith('.kwik'):
                basename, _ = op.splitext(basename)
            kwik_path = basename + '.kwik'
        self.kwik_path = kwik_path
        if basename is None:
            basename, _ = op.splitext(kwik_path)
        self.basename = basename

        # Find the .kwx filename.
        if kwx_path is None:
            basename, _ = op.splitext(kwik_path)
            kwx_path = basename + '.kwx'
        self.kwx_path = kwx_path

    def create_empty(self):
        """Create empty `.kwik` and `.kwx` files."""
        for path in (self.kwik_path, self.kwx_path):
            assert not op.exists(path)
            with open_h5(path, 'w') as f:
                f.write_attr('/', 'kwik_version', 2)
                f.write_attr('/', 'name', self.basename)
                f.write_attr('/', 'creator_version',
                             'klusta ' + __version_git__)

    def set_metadata(self, path, **kwargs):
        """Set metadata fields in a HDF5 path."""
        assert isinstance(path, string_types)
        assert path
        with open_h5(self.kwik_path, 'a') as f:
            for key, value in kwargs.items():
                f.write_attr(path, key, value)

    def set_probe(self, probe):
        """Save a probe dictionary in the file."""
        with open_h5(self.kwik_path, 'a') as f:
            probe = probe['channel_groups']
            for group, d in probe.items():
                group = int(group)
                channels = np.array(list(d['channels']), dtype=np.int32)

                # Write the channel order.
                f.write_attr('/channel_groups/{:d}'.format(group),
                             'channel_order', channels)

                # Write the probe adjacency graph.
                graph = d.get('graph', [])
                graph = np.array(graph, dtype=np.int32)
                f.write_attr('/channel_groups/{:d}'.format(group),
                             'adjacency_graph', graph)

                # Write the channel positions.
                positions = d.get('geometry', {})
                for channel in channels:
                    channel = int(channel)
                    # Get the channel position.
                    if channel in positions:
                        position = positions[channel]
                    else:
                        # Default position.
                        position = (0, channel)
                    path = '/channel_groups/{:d}/channels/{:d}'.format(
                        group, channel)

                    f.write_attr(path, 'name', str(channel))

                    position = np.array(position, dtype=np.float32)
                    f.write_attr(path, 'position', position)

    def add_spikes(self,
                   group=None,
                   spike_samples=None,
                   spike_recordings=None,
                   masks=None,
                   features=None,
                   n_channels=None,
                   n_features=None,
                   ):
        """Add spikes in the file.

        Parameters
        ----------

        group : int
            Channel group.
        spike_samples : ndarray
            The spike times in number of samples.
        spike_recordings : ndarray (optional)
            The recording indices of every spike.
        masks : ndarray or list of ndarrays
            The masks.
        features : ndarray or list of ndarrays
            The features.

        Note
        ----

        The features and masks can be passed as lists or generators of
        (memmapped) data chunks in order to avoid loading the entire arrays
        in RAM.

        """
        assert group >= 0
        assert n_channels >= 0
        assert n_features >= 0

        if spike_samples is None:
            return
        if isinstance(spike_samples, list):
            spike_samples = _concat(spike_samples)
        spike_samples = np.asarray(spike_samples, dtype=np.float64).ravel()
        n_spikes = len(spike_samples)
        if spike_recordings is None:
            spike_recordings = np.zeros(n_spikes, dtype=np.int32)
        spike_recordings = spike_recordings.ravel()

        # Add spikes in the .kwik file.
        assert op.exists(self.kwik_path)
        with open_h5(self.kwik_path, 'a') as f:
            # This method can only be called once.
            if '/channel_groups/{:d}/spikes/time_samples'.format(group) in f:
                raise RuntimeError("Spikes have already been added to this "
                                   "dataset.")
            time_samples = spike_samples.astype(np.uint64)
            frac = ((spike_samples - time_samples) * 255).astype(np.uint8)
            f.write('/channel_groups/{:d}/spikes/time_samples'.format(group),
                    time_samples)
            f.write('/channel_groups/{}/spikes/time_fractional'.format(group),
                    frac)
            f.write('/channel_groups/{:d}/spikes/recording'.format(group),
                    spike_recordings)

        if masks is None and features is None:
            return
        # Add features and masks in the .kwx file.
        assert masks is not None
        assert features is not None

        # Determine the shape of the features_masks array.
        shape = (n_spikes, n_channels * n_features, 2)

        def transform_f(f):
            return f[...].reshape((-1, n_channels * n_features))

        def transform_m(m):
            return np.repeat(m[...], n_features, axis=1)

        assert op.exists(self.kwx_path)
        with open_h5(self.kwx_path, 'a') as f:
            fm = f.write('/channel_groups/{:d}/features_masks'.format(group),
                         shape=shape, dtype=np.float32)

            # Write the features and masks in one block.
            if (isinstance(features, np.ndarray) and
                    isinstance(masks, np.ndarray)):
                fm[:, :, 0] = transform_f(features)
                fm[:, :, 1] = transform_m(masks)
            # Write the features and masks chunk by chunk.
            else:
                # Concatenate the features/masks chunks.
                fm_arrs = (np.dstack((transform_f(fet), transform_m(m)))
                           for (fet, m) in zip(features, masks)
                           if fet is not None and m is not None)
                _write_by_chunk(fm, fm_arrs)

    def add_spikes_after_detection(self, out):
        # Add the spikes in the `.kwik` and `.kwx` files.
        for group in out.groups:
            self.add_spikes(group=group,
                            spike_samples=out.spike_samples[group],
                            spike_recordings=out.recordings[group],
                            masks=out.masks[group],
                            features=out.features[group],
                            n_channels=out.n_channels_per_group[group],
                            n_features=out.n_features_per_channel,
                            )

    def add_recording(self, id=None, raw_path=None,
                      start_sample=None, sample_rate=None):
        """Add a recording.

        Parameters
        ----------

        id : int
            The recording id (0, 1, 2, etc.).
        raw_path : str
            Path to the file containing the raw data.
        start_sample : int
            The offset of the recording, in number of samples.
        sample_rate : float
            The sample rate of the recording

        """
        path = '/recordings/{:d}'.format(id)
        start_sample = int(start_sample)
        sample_rate = float(sample_rate)

        with open_h5(self.kwik_path, 'a') as f:
            f.write_attr(path, 'name', 'recording_{:d}'.format(id))
            f.write_attr(path, 'start_sample', start_sample)
            f.write_attr(path, 'sample_rate', sample_rate)
            f.write_attr(path, 'start_time', start_sample / sample_rate)
            if raw_path:
                if op.splitext(raw_path)[1] == '.kwd':
                    f.write_attr(path + '/raw', 'hdf5_path', raw_path)
                elif op.splitext(raw_path)[1] == '.dat':
                    f.write_attr(path + '/raw', 'dat_path', raw_path)

    def _add_recordings_from_dat(self, files, sample_rate=None,
                                 n_channels=None, dtype=None):
        start_sample = 0
        for i, filename in enumerate(files):
            # WARNING: different sample rates in recordings is not
            # supported yet.
            self.add_recording(id=i,
                               start_sample=start_sample,
                               sample_rate=sample_rate,
                               raw_path=filename,
                               )
            assert op.splitext(filename)[1] == '.dat'
            # Compute the offset for different recordings.
            start_sample += _dat_n_samples(filename,
                                           n_channels=n_channels,
                                           dtype=dtype)

    def _add_recordings_from_kwd(self, file, sample_rate=None):
        assert file.endswith('.kwd')
        start_sample = 0
        with open_h5(file, 'r') as f:
            recordings = f.children('/recordings')
            for recording in recordings:
                path = '/recordings/{}'.format(recording)
                if f.has_attr(path, 'sample_rate'):
                    sample_rate = f.read_attr(path, 'sample_rate')
                assert sample_rate > 0
                self.add_recording(id=int(recording),
                                   start_sample=start_sample,
                                   sample_rate=sample_rate,
                                   raw_path=file,
                                   )
                start_sample += f.read(path + '/data').shape[0]

    def add_cluster_group(self,
                          group=None,
                          id=None,
                          name=None,
                          clustering=None,
                          ):
        """Add a cluster group.

        Parameters
        ----------

        group : int
            The channel group.
        id : int
            The cluster group id.
        name : str
            The cluster group name.
        clustering : str
            The name of the clustering.

        """
        assert group >= 0
        cg_path = ('/channel_groups/{0:d}/'
                   'cluster_groups/{1:s}/{2:d}').format(group,
                                                        clustering,
                                                        id,
                                                        )
        with open_h5(self.kwik_path, 'a') as f:
            f.write_attr(cg_path, 'name', name)

    def add_clustering(self,
                       group=None,
                       name=None,
                       spike_clusters=None,
                       cluster_groups=None,
                       ):
        """Add a clustering.

        Parameters
        ----------

        group : int
            The channel group.
        name : str
            The clustering name.
        spike_clusters : ndarray
            The spike clusters assignements. This is `(n_spikes,)` array.
        cluster_groups : dict
            The cluster group of every cluster.

        """
        if cluster_groups is None:
            cluster_groups = {}
        path = '/channel_groups/{0:d}/spikes/clusters/{1:s}'.format(
            group, name)

        with open_h5(self.kwik_path, 'a') as f:
            assert not f.exists(path)

            # Save spike_clusters.
            spike_clusters = spike_clusters.astype(np.int32).ravel()
            f.write(path, spike_clusters)
            cluster_ids = _unique(spike_clusters)

            # Create cluster metadata.
            for cluster in cluster_ids:
                cluster_path = '/channel_groups/{0:d}/clusters/{1:s}/{2:d}'. \
                    format(group, name, cluster)

                # Default group: unsorted.
                cluster_group = cluster_groups.get(cluster, 3)
                f.write_attr(cluster_path, 'cluster_group', cluster_group)

            # Create cluster group metadata.
            for group_id, cg_name in _DEFAULT_GROUPS:
                self.add_cluster_group(id=group_id,
                                       name=cg_name,
                                       clustering=name,
                                       group=group,
                                       )


def create_kwik(prm_file=None,
                prm=None,
                kwik_path=None,
                output_dir=None,
                probe=None,
                overwrite=False,
                **kwargs):
    """Create a new Kwik dataset from a PRM file."""
    prm = prm or (_read_python(prm_file) if prm_file else {})
    if prm:
        assert 'spikedetekt' in prm
        assert 'traces' in prm
        sample_rate = prm['traces']['sample_rate']

    if 'sample_rate' in kwargs:
        sample_rate = kwargs['sample_rate']
    assert sample_rate > 0

    # Check output dir.
    output_dir = output_dir or ''
    if output_dir:
        _ensure_dir_exists(output_dir)

    # Ensure kwik path.
    kwik_path = kwik_path or op.join(output_dir,
                                     prm['experiment_name'] + '.kwik')
    kwik_path = op.realpath(kwik_path)
    assert kwik_path

    # Overwrite the kwik file if necessary.
    kwx_path = op.splitext(kwik_path)[0] + '.kwx'
    if op.exists(kwik_path):
        if overwrite:
            os.remove(kwik_path)
            os.remove(kwx_path)
        else:
            raise IOError("The `.kwik` file already exists. Please use "
                          "the `--overwrite` option.")

    # Ensure experiment name.
    experiment_name = op.basename(kwik_path)
    assert experiment_name

    # Default SpikeDetekt parameters.
    curdir = op.realpath(op.dirname(__file__))
    path = op.join(curdir, '../traces/default_settings.py')
    params = _read_python(path)['spikedetekt']
    # Update with PRM and user parameters.
    if prm:
        params['experiment_name'] = experiment_name
        params['prb_file'] = prm.get('prb_file', None)
        params.update(prm['spikedetekt'])
        params.update(prm['traces'])
    params.update(kwargs)

    # Ensure the probe file exists if it is required.
    probe = probe or load_probe(params['prb_file'])
    assert probe

    # KwikCreator.
    creator = KwikCreator(kwik_path)
    logger.info("Using kwik file `%s`.", kwik_path)
    # Create an empty kwik file if it doesn't exist yet.
    if not op.exists(kwik_path):
        creator.create_empty()
    creator.set_probe(probe)

    # Add the recordings.
    raw_data_files = params.get('raw_data_files', None)
    if isinstance(raw_data_files, string_types):
        if raw_data_files.endswith('.raw.kwd'):
            creator._add_recordings_from_kwd(raw_data_files,
                                             sample_rate=sample_rate,
                                             )
        else:
            raw_data_files = [raw_data_files]
    if isinstance(raw_data_files, list) and len(raw_data_files):
        # The dtype must be a string so that it can be serialized in HDF5.
        if not params.get('dtype', None):
            logger.warn("The `dtype` parameter is mandatory. Using a "
                        "default value of `int16` for now. Please update "
                        "your `.prm` file.")
            params['dtype'] = 'int16'
        assert 'dtype' in params and isinstance(params['dtype'], string_types)
        dtype = np.dtype(params['dtype'])
        assert dtype is not None

        # The number of channels in the .dat file *must* be specified.
        n_channels = params['n_channels']
        assert n_channels > 0
        creator._add_recordings_from_dat(raw_data_files,
                                         sample_rate=sample_rate,
                                         n_channels=n_channels,
                                         dtype=dtype,
                                         )

    creator.set_metadata('/application_data/spikedetekt',
                         **params)
    creator.set_metadata('/application_data/klustakwik2',
                         **prm.get('klustakwik2', {}))

    return kwik_path



# -*- coding: utf-8 -*-

"""Waveform extraction."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging

import numpy as np
from scipy.interpolate import interp1d

from ..utils import Bunch, _pad

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Waveform extractor from a connected component
#------------------------------------------------------------------------------

def _get_padded(data, start, end):
    """Return `data[start:end]` filling in with zeros outside array bounds

    Assumes that either `start<0` or `end>len(data)` but not both.

    """
    if start < 0 and end >= data.shape[0]:
        raise RuntimeError()
    if start < 0:
        start_zeros = np.zeros((-start, data.shape[1]),
                               dtype=data.dtype)
        return np.vstack((start_zeros, data[:end]))
    elif end > data.shape[0]:
        end_zeros = np.zeros((end - data.shape[0], data.shape[1]),
                             dtype=data.dtype)
        return np.vstack((data[start:], end_zeros))
    else:
        return data[start:end]


class WaveformExtractor(object):
    """Extract waveforms after data filtering and spike detection."""
    def __init__(self,
                 extract_before=None,
                 extract_after=None,
                 weight_power=None,
                 thresholds=None,
                 channels_per_group=None,
                 ):
        self._extract_before = extract_before
        self._extract_after = extract_after
        self._weight_power = weight_power if weight_power is not None else 1.
        self._thresholds = thresholds or {}
        self._channels_per_group = channels_per_group
        # mapping channel => channels in the shank
        self._dep_channels = {i: channels
                              for channels in channels_per_group.values()
                              for i in channels}
        self._channel_groups = {i: g
                                for g, channels in channels_per_group.items()
                                for i in channels}

    def _component(self, component, data=None, n_samples=None):
        comp_s = component[:, 0]  # shape: (component_size,)
        comp_ch = component[:, 1]  # shape: (component_size,)
        channel = comp_ch[0]
        if channel not in self._dep_channels:
            raise RuntimeError("Channel `{}` appears to be dead and should "
                               "have been excluded from the threshold "
                               "crossings.".format(channel))
        channels = self._dep_channels[channel]
        group = self._channel_groups[comp_ch[0]]

        # Get the temporal window around the waveform.
        s_min, s_max = (comp_s.min() - 3), (comp_s.max() + 4)
        s_min = max(s_min, 0)
        s_max = min(s_max, n_samples)
        assert s_min < s_max

        return Bunch(comp_s=comp_s,
                     comp_ch=comp_ch,
                     s_min=s_min,
                     s_max=s_max,
                     channels=channels,
                     group=group,
                     )

    def _normalize(self, x):
        x = np.asarray(x)
        tw = self._thresholds['weak']
        ts = self._thresholds['strong']
        return np.clip((x - tw) / (ts - tw), 0, 1)

    def _comp_wave(self, data_t, comp):
        comp_s, comp_ch = comp.comp_s, comp.comp_ch
        s_min, s_max = comp.s_min, comp.s_max
        nc = data_t.shape[1]
        # Data on weak threshold crossings. shape: (some_length, nc)
        wave = np.zeros((s_max - s_min, nc), dtype=data_t.dtype)
        # The sample where the peak is reached, on each channel.
        wave[comp_s - s_min, comp_ch] = data_t[comp_s, comp_ch]
        return wave

    def masks(self, data_t, wave, comp):
        nc = data_t.shape[1]
        channels = comp.channels
        comp_ch = comp.comp_ch
        s_min = comp.s_min

        # Binary mask. shape: (nc,)
        masks_bin = np.zeros(nc, dtype=np.bool)
        masks_bin[np.unique(comp_ch)] = 1

        # Find the peaks (relative to the start of the chunk). shape: (nc,)
        peaks = np.argmax(wave, axis=0) + s_min
        # Peak values on each channel. shape: (nc,)
        peaks_values = data_t[peaks, np.arange(0, nc)] * masks_bin

        # Compute the float masks.
        masks_float = self._normalize(peaks_values)
        # Keep shank channels.
        masks_float = masks_float[channels]
        return masks_float

    def spike_sample_aligned(self, wave, comp):
        s_min, s_max = comp.s_min, comp.s_max
        # Compute the fractional peak.
        wave_n = self._normalize(wave)
        wave_n_p = np.power(wave_n, self._weight_power)
        u = np.arange(s_max - s_min)[:, np.newaxis]
        # Spike aligned time relative to the start of the chunk.
        s_aligned = np.sum(wave_n_p * u) / np.sum(wave_n_p) + s_min
        return s_aligned

    def extract(self, data, s_aligned, channels=None):
        s = int(s_aligned)
        # Get block of given size around peak sample.
        waveform = _get_padded(data,
                               s - self._extract_before - 1,
                               s + self._extract_after + 2)
        return waveform[:, channels]  # Keep shank channels.

    def align(self, waveform, s_aligned):
        s = int(s_aligned)
        sb, sa = self._extract_before, self._extract_after
        # Perform interpolation around the fractional peak.
        old_s = np.arange(s - sb - 1, s + sa + 2)
        new_s = np.arange(s - sb + 0, s + sa + 0) + (s_aligned - s)
        try:
            f = interp1d(old_s, waveform, bounds_error=True,
                         kind='cubic', axis=0)
        except ValueError:
            logger.warn("Interpolation error at time {0:d}".format(s))
            return waveform
        return f(new_s)

    def set_thresholds(self, **kwargs):
        self._thresholds.update(kwargs)

    def __call__(self, component=None, data=None, data_t=None):
        assert data.shape == data_t.shape
        comp = self._component(component,
                               data=data,
                               n_samples=data_t.shape[0],
                               )
        channels = comp.channels

        wave = self._comp_wave(data_t, comp)
        masks = self.masks(data_t, wave, comp)
        s_aligned = self.spike_sample_aligned(wave, comp)

        waveform_unaligned = self.extract(data, s_aligned, channels=channels)
        waveform_aligned = self.align(waveform_unaligned, s_aligned)

        assert waveform_aligned.ndim == 2
        assert masks.ndim == 1
        assert waveform_aligned.shape[1] == masks.shape[0]

        return comp.group, s_aligned, waveform_aligned, masks


#------------------------------------------------------------------------------
# Waveform loader from traces (used in the manual sorting GUI)
#------------------------------------------------------------------------------

def _before_after(n_samples):
    """Get the number of samples before and after."""
    if not isinstance(n_samples, (tuple, list)):
        before = n_samples // 2
        after = n_samples - before
    else:
        assert len(n_samples) == 2
        before, after = n_samples
        n_samples = before + after
    assert before >= 0
    assert after >= 0
    assert before + after == n_samples
    return before, after


def _slice(index, n_samples, margin=None):
    """Return a waveform slice."""
    if margin is None:
        margin = (0, 0)
    assert isinstance(n_samples, (tuple, list))
    assert len(n_samples) == 2
    before, after = n_samples
    assert isinstance(margin, (tuple, list))
    assert len(margin) == 2
    margin_before, margin_after = margin
    before += margin_before
    after += margin_after
    index = int(index)
    before = int(before)
    after = int(after)
    return slice(max(0, index - before), index + after, None)


class WaveformLoader(object):
    """Load waveforms from filtered or unfiltered traces."""

    def __init__(self,
                 traces=None,
                 offset=0,
                 filter=None,
                 filter_margin=0,
                 n_samples=None,
                 channels=None,
                 scale_factor=None,
                 dc_offset=None,
                 ):
        self.n_samples_trace = 0
        # A (possibly memmapped) array-like structure with traces.
        if traces is not None:
            self.traces = traces
        else:
            self._traces = None
        self.dtype = np.float32
        # Scale factor for the loaded waveforms.
        self._scale_factor = scale_factor
        self._dc_offset = dc_offset
        # Offset of the traces: time (in samples) of the first trace sample.
        self._offset = int(offset)
        # List of channels to use when loading the waveforms.
        self._channels = channels
        # A filter function that takes a (n_samples, n_channels) array as
        # input.
        self._filter = filter
        # Number of samples to return, can be an int or a
        # tuple (before, after).
        if n_samples is None:
            raise ValueError("'n_samples' must be specified.")
        self.n_samples_before_after = _before_after(n_samples)
        self.n_samples_waveforms = sum(self.n_samples_before_after)
        # Number of additional samples to use for filtering.
        self._filter_margin = _before_after(filter_margin)
        # Number of samples in the extracted raw data chunk.
        self._n_samples_extract = (self.n_samples_waveforms +
                                   sum(self._filter_margin))

    @property
    def traces(self):
        """Raw traces."""
        return self._traces

    @traces.setter
    def traces(self, value):
        self.n_samples_trace, self.n_channels_traces = value.shape
        self._traces = value

    @property
    def channels(self):
        """List of channels."""
        return self._channels

    @channels.setter
    def channels(self, value):
        self._channels = value

    @property
    def n_channels_waveforms(self):
        """Number of channels kept for the waveforms."""
        if self._channels is not None:
            return len(self._channels)
        else:
            return self.n_channels_traces

    def _load_at(self, time):
        """Load a waveform at a given time."""
        time = int(time)
        time_o = time - self._offset
        ns = self.n_samples_trace
        if not (0 <= time_o < ns):
            raise ValueError("Invalid time {0:d}/{1:d}.".format(time_o,
                                                                ns))
        slice_extract = _slice(time_o,
                               self.n_samples_before_after,
                               self._filter_margin)
        extract = self._traces[slice_extract]

        # Pad the extracted chunk if needed.
        if slice_extract.start <= 0:
            extract = _pad(extract, self._n_samples_extract, 'left')
        elif slice_extract.stop >= ns - 1:
            extract = _pad(extract, self._n_samples_extract, 'right')

        assert extract.shape[0] == self._n_samples_extract

        # Filter the waveforms.
        # TODO: do the filtering in a vectorized way for higher performance.
        if self._filter is not None:
            waveforms = self._filter(extract)
        else:
            waveforms = extract

        # Remove the margin.
        margin_before, margin_after = self._filter_margin
        if margin_after > 0:
            assert margin_before >= 0
            waveforms = waveforms[margin_before:-margin_after, :]

        # Make a subselection with the specified channels.
        if self._channels is not None:
            out = waveforms[..., self._channels]
        else:
            out = waveforms

        assert out.shape == (self.n_samples_waveforms,
                             self.n_channels_waveforms)
        return out

    def __getitem__(self, item):
        """Load waveforms."""
        if isinstance(item, slice):
            raise NotImplementedError("Indexing with slices is not "
                                      "implemented yet.")
        if not isinstance(item, (tuple, list, np.ndarray)):
            item = [item]
        # Ensure a list of time samples are being requested.
        spikes = np.asarray(item)
        try:
            n_spikes = len(spikes)
        except TypeError:
            # HACK: handle the case where item is an unsized np.ndarray
            # (weird sort of scalar).
            spikes = np.asarray([item])
            n_spikes = len(spikes)
        # Initialize the array.
        # TODO: int16
        shape = (n_spikes, self.n_samples_waveforms,
                 self.n_channels_waveforms)
        # No traces: return null arrays.
        if self.n_samples_trace == 0:
            return np.zeros(shape, dtype=self.dtype)
        waveforms = np.empty(shape, dtype=self.dtype)
        # Load all spikes.
        for i, time in enumerate(spikes):
            try:
                waveforms[i, ...] = self._load_at(time)
            except ValueError as e:
                logger.warn("Error while loading waveform: {0}".format(str(e)))
        if self._dc_offset:
            waveforms -= self._dc_offset
        if self._scale_factor:
            waveforms *= self._scale_factor
        return waveforms


class SpikeLoader(object):
    """Translate selection with spike ids into selection with
    absolute times."""
    def __init__(self, waveforms, spike_samples):
        self._spike_samples = spike_samples
        # waveforms is a WaveformLoader instance
        self._waveforms = waveforms
        self.dtype = waveforms.dtype
        self.shape = (len(spike_samples),
                      waveforms.n_samples_waveforms,
                      waveforms.n_channels_waveforms)

    def __getitem__(self, item):
        times = self._spike_samples[item]
        return self._waveforms[times]

    def __len__(self):
        return len(self._spike_samples)



# -*- coding: utf-8 -*-

"""Test launcher."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os.path as op
from textwrap import dedent

from pytest import fixture

from ..launch import klusta
from ..datasets import download_test_data
from ..kwik.model import KwikModel


#------------------------------------------------------------------------------
# Fixtures
#------------------------------------------------------------------------------

@fixture
def prm_path_real(tempdir):
    dat_path = download_test_data('test-32ch-10s.dat')
    prm = dedent("""
        prb_file = '1x32_buzsaki'
        experiment_name = 'test_real'
        traces = dict(
            raw_data_files = [r"%s"],
            voltage_gain = 10.,
            sample_rate = 20000,
            n_channels = 32,
            dtype = 'int16',
        )
        spikedetekt = {}
        klustakwik2 = dict(num_starting_clusters=10)
        """ % dat_path)
    path = op.join(tempdir, 'params.prm')
    with open(path, 'w') as f:
        f.write(prm)
    return path


@fixture
def prm_path_shanks(tempdir):
    dat_path = download_test_data('test-32ch-10s.dat')

    # Create the PRB.
    prb = dedent("""
                 channel_groups = {
                    0: {'channels': [0, 2],
                        'graph': [[0, 2]],
                        },
                    1: {'channels': [1, 4],
                        'graph': [[1, 4]],
                        }
                 }
                 """)
    path = op.join(tempdir, 'probe.prb')
    with open(path, 'w') as f:
        f.write(prb)

    # Create the PRM.
    prm = dedent("""
        prb_file = r"%s"
        experiment_name = 'test_shanks'
        traces = dict(
            raw_data_files = [r"%s"],
            sample_rate = 20000,
            n_channels = 32,
            dtype = 'int16',
        )
        spikedetekt = {}
        klustakwik2 = dict(num_starting_clusters=20)
        """ % (path, dat_path))
    path = op.join(tempdir, 'params.prm')
    with open(path, 'w') as f:
        f.write(prm)

    return path


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_launch_shanks(tempdir, prm_path_shanks):
    kwik_path = klusta(prm_path_shanks,
                       output_dir=tempdir,
                       )

    model = KwikModel(kwik_path)

    model.channel_group = 0
    model.describe()
    assert model.n_spikes > 0
    assert model.n_clusters > 0

    model.channel_group = 1
    model.describe()
    assert model.n_spikes > 0
    assert model.n_clusters > 0

    assert model.clustering_metadata['klustakwik2_num_starting_clusters'] == 20

    klusta(kwik_path)


def test_launch_real(tempdir, prm_path_real):
    kwik_path = klusta(prm_path_real,
                       interval=(0., 4.),
                       output_dir=tempdir,
                       )

    # Check.
    model = KwikModel(kwik_path)
    assert model.n_spikes >= 100
    assert model.n_clusters >= 2
    model.describe()

    # Cluster only with different params.
    path = op.join(tempdir, 'params.prm')
    with open(path, 'a') as f:
        f.write('klustakwik2 = dict(num_starting_clusters=20)')

    kwik_path = klusta(prm_path_real,
                       interval=(0., 4.),
                       output_dir=tempdir,
                       cluster_only=True,
                       )

    # Check that the new param has been taken into account.
    model = KwikModel(kwik_path)
    assert model.kk2_metadata['num_starting_clusters'] == 20



from contextlib import contextmanager
import logging
import os
import os.path as op
import sys

import numpy as np
from six import exec_, integer_types
from six.moves import StringIO


class Bunch(dict):
    """A dict with additional dot syntax."""
    def __init__(self, *args, **kwargs):
        super(Bunch, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def copy(self):
        return Bunch(super(Bunch, self).copy())


def _pad(arr, n, dir='right'):
    """Pad an array with zeros along the first axis.

    Parameters
    ----------

    n : int
        Size of the returned array in the first axis.
    dir : str
        Direction of the padding. Must be one 'left' or 'right'.

    """
    assert dir in ('left', 'right')
    if n < 0:
        raise ValueError("'n' must be positive: {0}.".format(n))
    elif n == 0:
        return np.zeros((0,) + arr.shape[1:], dtype=arr.dtype)
    n_arr = arr.shape[0]
    shape = (n,) + arr.shape[1:]
    if n_arr == n:
        assert arr.shape == shape
        return arr
    elif n_arr < n:
        out = np.zeros(shape, dtype=arr.dtype)
        if dir == 'left':
            out[-n_arr:, ...] = arr
        elif dir == 'right':
            out[:n_arr, ...] = arr
        assert out.shape == shape
        return out
    else:
        if dir == 'left':
            out = arr[-n:, ...]
        elif dir == 'right':
            out = arr[:n, ...]
        assert out.shape == shape
        return out


def _index_of(arr, lookup):
    """Replace scalars in an array by their indices in a lookup table.

    Implicitely assume that:

    * All elements of arr and lookup are non-negative integers.
    * All elements or arr belong to lookup.

    This is not checked for performance reasons.

    """
    # Equivalent of np.digitize(arr, lookup) - 1, but much faster.
    # TODO: assertions to disable in production for performance reasons.
    # TODO: np.searchsorted(lookup, arr) is faster on small arrays with large
    # values
    lookup = np.asarray(lookup, dtype=np.int32)
    m = (lookup.max() if len(lookup) else 0) + 1
    tmp = np.zeros(m + 1, dtype=np.int)
    # Ensure that -1 values are kept.
    tmp[-1] = -1
    if len(lookup):
        tmp[lookup] = np.arange(len(lookup))
    return tmp[arr]


def _is_integer(x):
    return isinstance(x, integer_types + (np.generic,))


def _as_tuple(item):
    """Ensure an item is a tuple."""
    if item is None:
        return None
    elif not isinstance(item, tuple):
        return (item,)
    else:
        return item


def _spikes_per_cluster(spike_ids, spike_clusters):
    """Return a dictionary {cluster: list_of_spikes}."""
    if not len(spike_ids):
        return {}
    rel_spikes = np.argsort(spike_clusters)
    abs_spikes = spike_ids[rel_spikes]
    spike_clusters = spike_clusters[rel_spikes]

    diff = np.empty_like(spike_clusters)
    diff[0] = 1
    diff[1:] = np.diff(spike_clusters)

    idx = np.nonzero(diff > 0)[0]
    clusters = spike_clusters[idx]

    spikes_in_clusters = {clusters[i]: np.sort(abs_spikes[idx[i]:idx[i + 1]])
                          for i in range(len(clusters) - 1)}
    spikes_in_clusters[clusters[-1]] = np.sort(abs_spikes[idx[-1]:])

    return spikes_in_clusters


def _read_python(path):
    path = op.realpath(op.expanduser(path))
    assert op.exists(path)
    with open(path, 'r') as f:
        contents = f.read()
    metadata = {}
    exec_(contents, {}, metadata)
    metadata = {k.lower(): v for (k, v) in metadata.items()}
    return metadata


def _unique(x):
    """Faster version of np.unique().

    This version is restricted to 1D arrays of non-negative integers.

    It is only faster if len(x) >> len(unique(x)).

    """
    if x is None or len(x) == 0:
        return np.array([], dtype=np.int64)
    # WARNING: only keep positive values.
    # cluster=-1 means "unclustered".
    x = np.asarray(x)
    x = x[x >= 0]
    bc = np.bincount(x)
    return np.nonzero(bc)[0]


def _dat_n_samples(filename, dtype=None, n_channels=None, offset=None):
    assert dtype is not None
    item_size = np.dtype(dtype).itemsize
    offset = offset if offset else 0
    n_samples = (op.getsize(filename) - offset) // (item_size * n_channels)
    assert n_samples >= 0
    return n_samples


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


@contextmanager
def captured_logging(name=None):
    buffer = StringIO()
    logger = logging.getLogger(name)
    handlers = logger.handlers
    for handler in logger.handlers:
        logger.removeHandler(handler)
    handler = logging.StreamHandler(buffer)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    yield buffer
    buffer.flush()
    logger.removeHandler(handler)
    for handler in handlers:
        logger.addHandler(handler)


def _save_arrays(path, arrays):
    """Save multiple arrays in a single file by concatenating them along
    the first axis.
    A second array is stored with the offsets.
    """
    assert path.endswith('.npy')
    path = op.splitext(path)[0]
    offsets = np.cumsum([arr.shape[0] for arr in arrays])
    if not len(arrays):
        return
    concat = np.concatenate(arrays, axis=0)
    np.save(path + '.npy', concat)
    np.save(path + '.offsets.npy', offsets)


def _load_arrays(path):
    assert path.endswith('.npy')
    if not op.exists(path):
        return []
    path = op.splitext(path)[0]
    concat = np.load(path + '.npy')
    offsets = np.load(path + '.offsets.npy')
    return np.split(concat, offsets[:-1], axis=0)


def _as_array(arr, dtype=None):
    """Convert an object to a numerical NumPy array.
    Avoid a copy if possible.
    """
    if isinstance(arr, np.ndarray) and dtype is None:
        return arr
    if isinstance(arr, integer_types + (float,)):
        arr = [arr]
    out = np.asarray(arr)
    if dtype is not None:
        if out.dtype != dtype:
            out = out.astype(dtype)
    return out


def _concatenate(arrs):
    if arrs is None:
        return
    arrs = [_as_array(arr) for arr in arrs if arr is not None]
    if not arrs:
        return
    return np.concatenate(arrs, axis=0)


def _ensure_dir_exists(path):
    if not op.exists(path):
        os.makedirs(path)


def _klusta_user_dir():
    """Return the absolute path to the klusta user directory."""
    home = op.expanduser("~")
    path = op.realpath(op.join(home, '.klusta'))
    return path


def _excerpt_step(n_samples, n_excerpts=None, excerpt_size=None):
    """Compute the step of an excerpt set as a function of the number
    of excerpts or their sizes."""
    assert n_excerpts >= 2
    step = max((n_samples - excerpt_size) // (n_excerpts - 1),
               excerpt_size)
    return step


def chunk_bounds(n_samples, chunk_size, overlap=0):
    """Return chunk bounds.
    Chunks have the form:
        [ overlap/2 | chunk_size-overlap | overlap/2 ]
        s_start   keep_start           keep_end     s_end
    Except for the first and last chunks which do not have a left/right
    overlap.
    This generator yields (s_start, s_end, keep_start, keep_end).
    """
    s_start = 0
    s_end = chunk_size
    keep_start = s_start
    keep_end = s_end - overlap // 2
    yield s_start, s_end, keep_start, keep_end

    while s_end - overlap + chunk_size < n_samples:
        s_start = s_end - overlap
        s_end = s_start + chunk_size
        keep_start = keep_end
        keep_end = s_end - overlap // 2
        if s_start < s_end:
            yield s_start, s_end, keep_start, keep_end

    s_start = s_end - overlap
    s_end = n_samples
    keep_start = keep_end
    keep_end = s_end
    if s_start < s_end:
        yield s_start, s_end, keep_start, keep_end


def excerpts(n_samples, n_excerpts=None, excerpt_size=None):
    """Yield (start, end) where start is included and end is excluded."""
    assert n_excerpts >= 2
    step = _excerpt_step(n_samples,
                         n_excerpts=n_excerpts,
                         excerpt_size=excerpt_size)
    for i in range(n_excerpts):
        start = i * step
        if start >= n_samples:
            break
        end = min(start + excerpt_size, n_samples)
        yield start, end


def data_chunk(data, chunk, with_overlap=False):
    """Get a data chunk."""
    assert isinstance(chunk, tuple)
    if len(chunk) == 2:
        i, j = chunk
    elif len(chunk) == 4:
        if with_overlap:
            i, j = chunk[:2]
        else:
            i, j = chunk[2:]
    else:
        raise ValueError("'chunk' should have 2 or 4 elements, "
                         "not {0:d}".format(len(chunk)))
    return data[i:j, ...]


def get_excerpts(data, n_excerpts=None, excerpt_size=None):
    assert n_excerpts is not None
    assert excerpt_size is not None
    if len(data) < n_excerpts * excerpt_size:
        return data
    elif n_excerpts == 0:
        return data[:0]
    elif n_excerpts == 1:
        return data[:excerpt_size]
    out = np.concatenate([data_chunk(data, chunk)
                          for chunk in excerpts(len(data),
                                                n_excerpts=n_excerpts,
                                                excerpt_size=excerpt_size)])
    assert len(out) <= n_excerpts * excerpt_size
    return out



# -*- coding: utf-8 -*-

"""The KwikModel class manages in-memory structures and Kwik file open/save."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from collections import defaultdict
import os
import os.path as op
from random import randint

import logging
import numpy as np
from six import string_types, integer_types

from ..traces.waveform import WaveformLoader, SpikeLoader
from ..traces.filter import bandpass_filter, apply_filter
from ..utils import _unique, _as_tuple, _read_python
from .h5 import open_h5, File, _mmap_h5
from .mea import MEA

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# PartialArray
# -----------------------------------------------------------------------------

class PartialArray(object):
    """Proxy to a view of an array, allowing selection along the first
    dimensions and fixing the trailing dimensions."""
    def __init__(self, arr, trailing_index=None, reshape=None):
        self._arr = arr
        self._trailing_index = _as_tuple(trailing_index)
        self.shape = _partial_shape(arr.shape, self._trailing_index)
        self.dtype = arr.dtype
        self.ndim = len(self.shape)
        # We can reshape selections dynamically.
        # Use to reshape the 2D features array into a 3D array.
        self.reshape = reshape
        if reshape and reshape[0] == -1:
            self.shape = (self.shape[0],) + self.reshape[1:]

    def __getitem__(self, item):
        if self._trailing_index is None:
            out = self._arr[item]
        else:
            item = _as_tuple(item)
            k = len(item)
            n = len(self._arr.shape)
            t = len(self._trailing_index)
            if k < (n - t):
                item += (slice(None, None, None),) * (n - k - t)
            item += self._trailing_index
            if len(item) != n:
                raise ValueError("The array selection is invalid: "
                                 "{0}".format(str(item)))
            out = self._arr[item]
        if self.reshape:
            # Avoid error when the reshaped array contains extra dimensions.
            n = np.prod(self.reshape[1:])
            return out[..., :n].reshape(self.reshape)
        else:
            return out

    def __len__(self):
        return self.shape[0]


class ConcatenatedArrays(object):
    """This object represents a concatenation of several memory-mapped
    arrays."""
    def __init__(self, arrs, cols=None):
        assert isinstance(arrs, list)
        self.arrs = arrs
        # Reordering of the columns.
        self.cols = cols
        self.offsets = np.concatenate([[0], np.cumsum([arr.shape[0]
                                                       for arr in arrs])],
                                      axis=0)
        self.dtype = arrs[0].dtype if arrs else None

    @property
    def shape(self):
        ncols = (len(self.cols) if self.cols is not None
                 else self.arrs[0].shape[1])
        return (self.offsets[-1], ncols)

    def _get_recording(self, index):
        """Return the recording that contains a given index."""
        assert index >= 0
        recs = np.nonzero((index - self.offsets[:-1]) >= 0)[0]
        if len(recs) == 0:
            # If the index is greater than the total size,
            # return the last recording.
            return len(self.arrs) - 1
        # Return the last recording such that the index is greater than
        # its offset.
        return recs[-1]

    def __getitem__(self, item):
        cols = self.cols if self.cols is not None else slice(None, None, None)
        # Get the start and stop indices of the requested item.
        start, stop = _start_stop(item)
        # Return the concatenation of all arrays.
        if start is None and stop is None:
            return np.concatenate(self.arrs, axis=0)[:, cols]
        if start is None:
            start = 0
        if stop is None:
            stop = self.offsets[-1]
        if stop < 0:
            stop = self.offsets[-1] + stop
        # Get the recording indices of the first and last item.
        rec_start = self._get_recording(start)
        rec_stop = self._get_recording(stop)
        assert 0 <= rec_start <= rec_stop < len(self.arrs)
        # Find the start and stop relative to the arrays.
        start_rel = start - self.offsets[rec_start]
        stop_rel = stop - self.offsets[rec_stop]
        # Single array case.
        if rec_start == rec_stop:
            # Apply the rest of the index.
            return _fill_index(self.arrs[rec_start][start_rel:stop_rel],
                               item)[:, cols]
        chunk_start = self.arrs[rec_start][start_rel:]
        chunk_stop = self.arrs[rec_stop][:stop_rel]
        # Concatenate all chunks.
        l = [chunk_start]
        if rec_stop - rec_start >= 2:
            logger.warn("Loading a full virtual array: this might be slow "
                        "and something might be wrong.")
            l += [self.arrs[r][...] for r in range(rec_start + 1,
                                                   rec_stop)]
        l += [chunk_stop]
        # Apply the rest of the index.
        return _fill_index(np.concatenate(l, axis=0), item)[:, cols]

    def __len__(self):
        return self.shape[0]


class VirtualMappedArray(object):
    """A virtual mapped array that yields null arrays to any selection."""
    def __init__(self, shape, dtype, fill=0):
        self.shape = shape
        self.dtype = dtype
        self.ndim = len(self.shape)
        self._fill = fill

    def __getitem__(self, item):
        if isinstance(item, integer_types):
            return self._fill * np.ones(self.shape[1:], dtype=self.dtype)
        else:
            assert not isinstance(item, tuple)
            n = _len_index(item, max_len=self.shape[0])
            return self._fill * np.ones((n,) + self.shape[1:],
                                        dtype=self.dtype)

    def __len__(self):
        return self.shape[0]


#------------------------------------------------------------------------------
# Utility functions
#------------------------------------------------------------------------------

def _start_stop(item):
    """Find the start and stop indices of a __getitem__ item.

    This is used only by ConcatenatedArrays.

    Only two cases are supported currently:

    * Single integer.
    * Contiguous slice in the first dimension only.

    """
    if isinstance(item, tuple):
        item = item[0]
    if isinstance(item, slice):
        # Slice.
        if item.step not in (None, 1):
            return NotImplementedError()
        return item.start, item.stop
    elif isinstance(item, (list, np.ndarray)):
        # List or array of indices.
        return np.min(item), np.max(item)
    else:
        # Integer.
        return item, item + 1


def _len_index(item, max_len=0):
    """Return the expected length of the output of __getitem__(item)."""
    if isinstance(item, (list, np.ndarray)):
        return len(item)
    elif isinstance(item, slice):
        stop = item.stop or max_len
        start = item.start or 0
        step = item.step or 1
        start = np.clip(start, 0, stop)
        assert 0 <= start <= stop
        return 1 + ((stop - 1 - start) // step)
    else:
        return 1


def _fill_index(arr, item):
    if isinstance(item, tuple):
        item = (slice(None, None, None),) + item[1:]
        return arr[item]
    else:
        return arr


def _partial_shape(shape, trailing_index):
    """Return the shape of a partial array."""
    if shape is None:
        shape = ()
    if trailing_index is None:
        trailing_index = ()
    trailing_index = _as_tuple(trailing_index)
    # Length of the selection items for the partial array.
    len_item = len(shape) - len(trailing_index)
    # Array for the trailing dimensions.
    _arr = np.empty(shape=shape[len_item:])
    try:
        trailing_arr = _arr[trailing_index]
    except IndexError:
        raise ValueError("The partial shape index is invalid.")
    return shape[:len_item] + trailing_arr.shape


def _concatenate_virtual_arrays(arrs, cols=None):
    """Return a virtual concatenate of several NumPy arrays."""
    n = len(arrs)
    if n == 0:
        return None
    return ConcatenatedArrays(arrs, cols)


def _dat_n_samples(filename, dtype=None, n_channels=None):
    assert dtype is not None
    item_size = np.dtype(dtype).itemsize
    n_samples = op.getsize(filename) // (item_size * n_channels)
    assert n_samples >= 0
    return n_samples


def _dat_to_traces(dat_path, n_channels, dtype):
    assert dtype is not None
    assert n_channels is not None
    n_samples = _dat_n_samples(dat_path,
                               n_channels=n_channels,
                               dtype=dtype,
                               )
    return np.memmap(dat_path, dtype=dtype, shape=(n_samples, n_channels))


#------------------------------------------------------------------------------
# Kwik utility functions
#------------------------------------------------------------------------------

def _to_int_list(l):
    """Convert int strings to ints."""
    return [int(_) for _ in l]


def _list_int_children(group):
    """Return the list of int children of a HDF5 group."""
    return sorted(_to_int_list(group.keys()))


def _list_channel_groups(kwik):
    """Return the list of channel groups in a kwik file."""
    if 'channel_groups' in kwik:
        return _list_int_children(kwik['/channel_groups'])
    else:
        return []


def _list_recordings(kwik):
    """Return the list of recordings in a kwik file."""
    if '/recordings' in kwik:
        recordings = _list_int_children(kwik['/recordings'])
    else:
        recordings = []
    # TODO: return a dictionary of recordings instead of a list of recording
    # ids.
    # return {rec: Bunch({
    #     'start': kwik['/recordings/{0}'.format(rec)].attrs['start_sample']
    # }) for rec in recordings}
    return recordings


def _list_channels(kwik, channel_group=None):
    """Return the list of channels in a kwik file."""
    assert isinstance(channel_group, integer_types)
    path = '/channel_groups/{0:d}/channels'.format(channel_group)
    if path in kwik:
        channels = _list_int_children(kwik[path])
        return channels
    else:
        return []


def _list_clusterings(kwik, channel_group=None):
    """Return the list of clusterings in a kwik file."""
    if channel_group is None:
        raise RuntimeError("channel_group must be specified when listing "
                           "the clusterings.")
    assert isinstance(channel_group, integer_types)
    path = '/channel_groups/{0:d}/clusters'.format(channel_group)
    if path not in kwik:
        return []
    clusterings = sorted(kwik[path].keys())
    # Ensure 'main' is the first if it exists.
    if 'main' in clusterings:
        clusterings.remove('main')
        clusterings = ['main'] + clusterings
    return clusterings


def _concatenate_spikes(spikes, recs, offsets):
    """Concatenate spike samples belonging to consecutive recordings."""
    assert offsets is not None
    spikes = np.asarray(spikes)
    offsets = np.asarray(offsets)
    recs = np.asarray(recs)
    return (spikes + offsets[recs]).astype(np.int64)


def _create_cluster_group(f, group_id, name,
                          clustering=None,
                          channel_group=None,
                          write_color=True,
                          ):
    cg_path = ('/channel_groups/{0:d}/'
               'cluster_groups/{1:s}/{2:d}').format(channel_group,
                                                    clustering,
                                                    group_id,
                                                    )
    kv_path = cg_path + '/application_data/klustaviewa'
    f.write_attr(cg_path, 'name', name)
    if write_color:
        f.write_attr(kv_path, 'color', randint(2, 10))


def _create_clustering(f, name,
                       channel_group=None,
                       spike_clusters=None,
                       cluster_groups=None,
                       ):
    if cluster_groups is None:
        cluster_groups = {}
    assert isinstance(f, File)
    path = '/channel_groups/{0:d}/spikes/clusters/{1:s}'.format(channel_group,
                                                                name)
    assert not f.exists(path)

    # Save spike_clusters.
    f.write(path, spike_clusters.astype(np.int32))

    cluster_ids = _unique(spike_clusters)

    # Create cluster metadata.
    for cluster in cluster_ids:
        cluster_path = '/channel_groups/{0:d}/clusters/{1:s}/{2:d}'.format(
            channel_group, name, cluster)
        kv_path = cluster_path + '/application_data/klustaviewa'

        # Default group: unsorted.
        cluster_group = cluster_groups.get(cluster, 3)
        f.write_attr(cluster_path, 'cluster_group', cluster_group)
        f.write_attr(kv_path, 'color', randint(2, 10))

    # Create cluster group metadata.
    for group_id, cg_name in _DEFAULT_GROUPS:
        _create_cluster_group(f, group_id, cg_name,
                              clustering=name,
                              channel_group=channel_group,
                              )


def list_kwik(folders):
    """Return the list of Kwik files found in a list of folders."""
    ret = []
    for d in folders:
        for root, dirs, files in os.walk(os.path.expanduser(d)):
            for f in files:
                if f.endswith(".kwik"):
                    ret.append(os.path.join(root, f))
    return ret


def _open_h5_if_exists(kwik_path, file_type, mode=None):
    basename, ext = op.splitext(kwik_path)
    path = '{basename}.{ext}'.format(basename=basename, ext=file_type)
    return open_h5(path, mode=mode) if op.exists(path) else None


def _read_traces(kwik, dtype=None, n_channels=None):
    kwd_path = None
    dat_path = None

    dtype = kwik.read_attr('/application_data/spikedetekt/', 'dtype', None)
    if dtype:
        dtype = np.dtype(dtype)

    n_channels = kwik.read_attr('/application_data/spikedetekt/',
                                'n_channels',
                                None)
    # HACK: old format.
    nchannels = kwik.read_attr('/application_data/spikedetekt/',
                               'nchannels',
                               None)
    n_channels = n_channels or nchannels
    if n_channels:
        n_channels = int(n_channels)

    if '/recordings' not in kwik:
        return (None, None)
    recordings = kwik.children('/recordings')
    traces = []
    opened_files = []

    # HACK when there is no recordings: find a .dat file with the same
    # base name in the current directory.
    if not recordings:
        name = op.splitext(op.basename(kwik.filename))[0]
        p = op.join(op.dirname(op.realpath(kwik.filename)), name + '.dat')
        if op.exists(p):
            logger.debug("Loading traces: %s", p)
            dat = _dat_to_traces(p, dtype=dtype or 'int16',
                                 n_channels=n_channels)
            traces.append(dat)
            opened_files.append(dat)

    for recording in recordings:
        # Is there a path specified to a .raw.kwd file which exists in
        # [KWIK]/recordings/[X]/raw? If so, open it.
        path = '/recordings/{}/raw'.format(recording)
        if kwik.has_attr(path, 'hdf5_path'):
            kwd_path = kwik.read_attr(path, 'hdf5_path')[:-8]
            kwd = _open_h5_if_exists(kwd_path, 'raw.kwd')
            if kwd is None:
                logger.debug("%s not found, trying same basename in KWIK dir",
                             kwd_path)
            else:
                logger.debug("Loading traces: %s", kwd_path)
                traces.append(kwd.read('/recordings/{}/data'
                              .format(recording)))
                opened_files.append(kwd)
                continue
        # Is there a path specified to a .dat file which exists?
        elif kwik.has_attr(path, 'dat_path'):
            assert dtype is not None
            assert n_channels
            dat_path = kwik.read_attr(path, 'dat_path')
            if not op.exists(dat_path):
                logger.debug("%s not found, trying same basename in KWIK dir",
                             dat_path)
            else:
                logger.debug("Loading traces: %s", dat_path)
                dat = _dat_to_traces(dat_path, dtype=dtype,
                                     n_channels=n_channels)
                traces.append(dat)
                opened_files.append(dat)
                continue

        # Does a file exist with the same name in the current directory?
        # If so, open it.
        if kwd_path is not None:
            rel_path = str(op.basename(kwd_path))
            rel_path = op.join(op.dirname(op.realpath(kwik.filename)),
                               rel_path)
            kwd = _open_h5_if_exists(rel_path, 'raw.kwd')
            if kwd is None:
                logger.debug("%s not found, trying experiment basename in "
                             "KWIK dir", rel_path)
            else:
                logger.debug("Loading traces: %s", rel_path)
                traces.append(kwd.read('/recordings/{}/data'
                              .format(recording)))
                opened_files.append(kwd)
                continue
        elif dat_path is not None:
            rel_path = op.basename(dat_path)
            rel_path = op.join(op.dirname(op.realpath(kwik.filename)),
                               rel_path)
            if not op.exists(rel_path):
                logger.debug("%s not found, trying experiment basename in "
                             "KWIK dir", rel_path)
            else:
                logger.debug("Loading traces: %s", rel_path)
                dat = _dat_to_traces(rel_path, dtype=dtype,
                                     n_channels=n_channels)
                traces.append(dat)
                opened_files.append(dat)
                continue

        # Finally, is there a `raw.kwd` with the experiment basename in the
        # current directory? If so, open it.
        kwd = _open_h5_if_exists(kwik.filename, 'raw.kwd')
        if kwd is None:
            logger.warn("Could not find any data source for traces "
                        "(raw.kwd or .dat or .bin.). Waveforms and "
                        "traces will not be available.")
        else:
            logger.debug("Successfully loaded basename.raw.kwd in "
                         "same directory")
            traces.append(kwd.read('/recordings/{}/data'.format(recording)))
            opened_files.append(kwd)

    return traces, opened_files


_DEFAULT_GROUPS = [(0, 'Noise'),
                   (1, 'MUA'),
                   (2, 'Good'),
                   (3, 'Unsorted'),
                   ]


def _check_spikes_increasing(spike_samples):

    while not np.all(np.diff(spike_samples) >= 0):
        # NOTE: for an unknown reason, spike times are not
        # always strictly increasing in Kwik files. Mostly, there is
        # a difference of 1 sample, e.g. [..., 123456, 123455, 123490, ...]
        # We try to fix this specific problem, otherwise we raise an
        # error. The program makes the assumption that spike times
        # are always increasing. Otherwise, the CCG computation may fail.
        # There might be other issues as well.
        msg = "The spike times must be increasing. "
        spk = np.nonzero(np.diff(spike_samples) < 0)[0]
        msg += "The spurious spike ids are: %s (%d)" % (str(spk[:9].tolist()),
                                                        len(spk))
        logger.debug(msg)
        # logger.debug(str(spike_samples[spk[:10]].tolist()))
        # logger.debug(str(spike_samples[spk[:10]+1].tolist()))
        logger.debug("Trying a quick hack to fix the problem.")
        spike_samples[spk + 1] += 1
    return spike_samples


#------------------------------------------------------------------------------
# KwikModel class
#------------------------------------------------------------------------------

class KwikModel(object):
    """Holds data contained in a kwik file."""

    def __init__(self, kwik_path=None,
                 channel_group=None,
                 clustering=None,
                 waveform_filter=True,
                 ):
        self.name = None
        self._channel_group = None
        self._clustering = None

        # Initialize fields.
        self._spike_samples = None
        self._spike_times = None
        self._spike_clusters = None
        self._metadata = None
        self._kk2_metadata = None
        self._clustering = clustering or 'main'
        self._probe = None
        self._channels = []
        self._channel_order = None
        self._features = None
        self._features_masks = None
        self._masks = None
        self._waveforms = None
        self._cluster_metadata = None
        self._clustering_metadata = {}
        self._all_traces = None
        self._traces = None
        self._recording_offsets = None
        self._waveform_loader = None
        self._waveform_filter = waveform_filter
        self._opened_files = []

        # Open the experiment.
        self.kwik_path = kwik_path
        self.open(kwik_path,
                  channel_group=channel_group,
                  clustering=clustering)

    @property
    def path(self):
        return self.kwik_path

    # Internal properties and methods
    # -------------------------------------------------------------------------

    def _check_kwik_version(self):
        # This class only works with kwik version 2 for now.
        kwik_version = self._kwik.read_attr('/', 'kwik_version')
        if kwik_version != 2:
            raise IOError("The kwik version is %d != 2." % kwik_version)

    @property
    def _channel_groups_path(self):
        return '/channel_groups/{0:d}'.format(self._channel_group)

    @property
    def _spikes_path(self):
        return '{0:s}/spikes'.format(self._channel_groups_path)

    @property
    def _channels_path(self):
        return '{0:s}/channels'.format(self._channel_groups_path)

    @property
    def _clusters_path(self):
        return '{0:s}/clusters'.format(self._channel_groups_path)

    @property
    def _cluster_groups_path(self):
        return '{0:s}/cluster_groups'.format(self._channel_groups_path)

    def _cluster_path(self, cluster):
        return '{0:s}/{1:d}'.format(self._clustering_path, cluster)

    @property
    def _spike_clusters_path(self):
        return '{0:s}/clusters/{1:s}'.format(self._spikes_path,
                                             self._clustering)

    @property
    def _clustering_path(self):
        return '{0:s}/{1:s}'.format(self._clusters_path, self._clustering)

    # Loading and saving
    # -------------------------------------------------------------------------

    def _open_kwik_if_needed(self, mode=None):
        if not self._kwik.is_open():
            self._kwik.open(mode=mode)
            return True
        else:
            if mode is not None:
                self._kwik.mode = mode
            return False

    @property
    def n_samples_waveforms(self):
        return (self._metadata['extract_s_before'] +
                self._metadata['extract_s_after'])

    def _create_waveform_loader(self):
        """Create a waveform loader."""
        n_samples = (self._metadata['extract_s_before'],
                     self._metadata['extract_s_after'])
        order = self._metadata['filter_butter_order']
        rate = self._metadata['sample_rate']
        low = self._metadata['filter_low']
        high = self._metadata['filter_high_factor'] * rate
        b_filter = bandpass_filter(rate=rate,
                                   low=low,
                                   high=high,
                                   order=order)

        if self._metadata.get('waveform_filter', True):
            logger.debug("Enable waveform filter.")

            def the_filter(x, axis=0):
                return apply_filter(x, b_filter, axis=axis)

            filter_margin = order * 3
        else:
            logger.debug("Disable waveform filter.")
            the_filter = None
            filter_margin = 0

        dc_offset = self._metadata.get('waveform_dc_offset', None)
        scale_factor = self._metadata.get('waveform_scale_factor', None)
        self._waveform_loader = WaveformLoader(n_samples=n_samples,
                                               filter=the_filter,
                                               filter_margin=filter_margin,
                                               dc_offset=dc_offset,
                                               scale_factor=scale_factor,
                                               )

    def _update_waveform_loader(self):
        if self._traces is not None:
            self._waveform_loader.traces = self._traces
        else:
            self._waveform_loader.traces = np.zeros((0, self.n_channels),
                                                    dtype=np.float32)

        # Update the list of channels for the waveform loader.
        # UPDATE: no need to do this because channel_order is now
        # taken into account in the traces
        # self._waveform_loader.channels = self._channel_order

    def _create_cluster_metadata(self):
        self._cluster_metadata = defaultdict(lambda: 'unsorted')

    def _load_meta(self, name='spikedetekt'):
        """Load metadata from kwik file."""
        # Automatically load all metadata from spikedetekt group.
        path = '/application_data/{}/'.format(name)
        params = {}
        for attr in self._kwik.attrs(path):
            params[attr] = self._kwik.read_attr(path, attr)
        # Load the default spikedetekt parameters.
        curdir = op.realpath(op.dirname(__file__))
        path = op.join(curdir, '../traces/default_settings.py')
        dfs = _read_python(path).get(name, {})
        for name, default_value in dfs.items():
            if name not in params:
                params[name] = default_value
        return params

    def _load_probe(self):
        # Re-create the probe from the Kwik file.
        channel_groups = {}
        for group in self._channel_groups:
            cg_p = '/channel_groups/{:d}'.format(group)
            c_p = cg_p + '/channels'
            channels = self._kwik.read_attr(cg_p, 'channel_order')
            graph = self._kwik.read_attr(cg_p, 'adjacency_graph')
            positions = {
                channel: self._kwik.read_attr(c_p + '/' + str(channel),
                                              'position')
                for channel in channels
            }
            channel_groups[group] = {
                'channels': channels,
                'graph': graph,
                'geometry': positions,
            }
        probe = {'channel_groups': channel_groups}
        self._probe = MEA(probe=probe)

    def _load_recordings(self):
        # Load recordings.
        self._recordings = _list_recordings(self._kwik.h5py_file)
        # This will be updated later if a KWD file is present.
        self._recording_offsets = [0] * (len(self._recordings) + 1)

    def _load_channels(self):
        self._channels = np.array(_list_channels(self._kwik.h5py_file,
                                                 self._channel_group))
        self._channel_order = self._probe.channels
        if self._traces is not None:
            self._traces.cols = self._channel_order
        assert set(self._channel_order) <= set(self._channels)

    def _load_channel_groups(self, channel_group=None):
        self._channel_groups = _list_channel_groups(self._kwik.h5py_file)
        if channel_group is None and self._channel_groups:
            # Choose the default channel group if not specified.
            channel_group = self._channel_groups[0]
        # Load the channel group.
        self._channel_group = channel_group

    def _load_features_masks(self):

        # Load features masks.
        h5_path = '{0:s}/features_masks'.format(self._channel_groups_path)

        nfpc = self._metadata['n_features_per_channel']
        nc = len(self.channel_order)

        basename, ext = op.splitext(self.kwik_path)
        kwx_path = basename + '.kwx'

        if not op.exists(kwx_path):
            logger.warn("The `.kwx` file hasn't been found. "
                        "Features and masks won't be available.")
            return

        with open_h5(kwx_path, 'r') as f:
            if h5_path not in f:
                logger.debug("There are no features and masks in the "
                             "`.kwx` file.")
                self._features_masks = self._features = self._masks = None
                return

        # Now we are sure that the kwx file exists and contains the features
        # and masks.

        # NOTE: we bypass HDF5 when reading the features_masks array,
        # for performance reasons. We retrieve the offset in the file
        # and use NumPy memmap.
        # fm = self._kwx.read(path)
        fm = _mmap_h5(kwx_path, h5_path)
        self._features_masks = fm
        self._features = PartialArray(fm, 0, reshape=(-1, nc, nfpc))

        # This partial array simulates a (n_spikes, n_channels) array.
        self._masks = PartialArray(fm, (slice(0, nfpc * nc, nfpc), 1))
        assert self._masks.shape == (self.n_spikes, nc)

    def _load_spikes(self):
        # Load spike samples.
        path = '{0:s}/time_samples'.format(self._spikes_path)

        # Concatenate the spike samples from consecutive recordings.
        if path not in self._kwik:
            logger.debug("No spikes found.")
            self._spike_recordings = np.array([], dtype=np.int32)
            self._spike_samples = None
            self._recording_offsets = np.array([], dtype=np.int32)
            return
        _spikes = self._kwik.read(path)[:]
        self._spike_recordings = self._kwik.read(
            '{0:s}/recording'.format(self._spikes_path))[:]

        # If the recording offsets have not been set, we need to create
        # non-trivial ones to ensure that spike times are increasing.
        if np.all(np.array(self._recording_offsets) == 0):
            bounds = np.nonzero(np.diff(self._spike_recordings) > 0)[0]
            self._recording_offsets = np.cumsum(np.hstack(([0],
                                                           _spikes[bounds],
                                                           _spikes[-1])))
        # FIX: remove +1 offset at each recording.
        self._recording_offsets -= np.arange(len(self._recording_offsets))
        self._spike_samples = _concatenate_spikes(_spikes,
                                                  self._spike_recordings,
                                                  self._recording_offsets)
        # Check that spikes are increasing.
        _check_spikes_increasing(self._spike_samples)

    def _load_spike_clusters(self):
        self._spike_clusters = self._kwik.read(self._spike_clusters_path)[:]

    def _save_spike_clusters(self, spike_clusters):
        assert spike_clusters.shape == self._spike_clusters.shape
        assert spike_clusters.dtype == self._spike_clusters.dtype
        self._spike_clusters = spike_clusters
        sc = self._kwik.read(self._spike_clusters_path)
        sc[:] = spike_clusters

    def _load_clusterings(self, clustering=None):
        # Once the channel group is loaded, list the clusterings.
        self._clusterings = _list_clusterings(self._kwik.h5py_file,
                                              self.channel_group)
        # Choose the first clustering (should always be 'main').
        if clustering is None and self.clusterings:
            clustering = self.clusterings[0]
        # Load the specified clustering.
        self._clustering = clustering

    @property
    def _cluster_groups_mapping(self):
        """Mapping {cluster_group_id: name}."""
        path = '{}/{}'.format(self._cluster_groups_path,
                              self._clustering)
        cluster_groups = sorted(map(int, self._kwik.groups(path)))
        mapping = {}
        for cluster_group in cluster_groups:
            name = self._kwik.read_attr('{}/{}'.format(path, cluster_group),
                                        'name')
            mapping[cluster_group] = name
        return mapping

    def _load_cluster_groups(self):
        clusters = self._kwik.groups(self._clustering_path)
        clusters = [int(cluster) for cluster in clusters]
        # NOTE: inverse mapping group name ==> group_number
        mapping = {a: b.lower() if b else str(b)
                   for a, b in self._cluster_groups_mapping.items()}
        imapping = {b: a for a, b in mapping.items()}
        for cluster in clusters:
            path = self._cluster_path(cluster)
            # NOTE: The group id is saved in the Kwik file. However,
            # it might happen that the Kwik file was saved with a bug that
            # replaced the group id by its name. Here we ensure that we
            # have a group id.
            group_id = self._kwik.read_attr(path, 'cluster_group')
            if isinstance(group_id, string_types):
                group_id = group_id.lower()
            # HACK: for an unknown reason, group_id might sometimes be an array
            if isinstance(group_id, np.ndarray):  # pragma: no cover
                group_id = group_id[0]
            group_id = int(imapping.get(group_id, group_id))
            # Get the group name.
            group = mapping.get(group_id, 'group_%d' % group_id).lower()
            assert group and isinstance(group, string_types)
            self._cluster_metadata[cluster] = group

    def _save_cluster_groups(self, cluster_groups):
        assert isinstance(cluster_groups, dict)
        # NOTE: inverse mapping group name ==> group_number
        mapping = {a: b.lower() if b else str(b)
                   for a, b in self._cluster_groups_mapping.items()}
        imapping = {b: a for a, b in mapping.items()}
        for cluster, group in cluster_groups.items():
            # Find the group and the group name.
            if isinstance(group, integer_types):
                group_id = group
                assert group_id in mapping
                group = mapping.get(group_id)
            elif group in (None, 'None'):
                group_id = imapping.get(group)
            elif isinstance(group, string_types):
                group = group.lower()
                assert group in imapping
                group_id = imapping.get(group)
            elif group is None:
                group_id = 3
                group = mapping.get(group_id)
            assert group_id is not None
            assert group_id >= 0
            assert group is not None
            assert isinstance(group, string_types)
            path = self._cluster_path(cluster)
            self._kwik.write_attr(path, 'cluster_group', group_id)
            self._cluster_metadata[cluster] = group

    def _load_clustering_metadata(self):
        attrs = self._kwik.attrs(self._clustering_path)
        metadata = {}
        for attr in attrs:
            try:
                metadata[attr] = self._kwik.read_attr(self._clustering_path,
                                                      attr)
            except OSError:
                logger.debug("Error when reading `%s:%s`.",
                             self._clustering_path, attr)
        self._clustering_metadata = metadata

    def _save_clustering_metadata(self, metadata):
        if not metadata:
            return
        assert isinstance(metadata, dict)
        for name, value in metadata.items():
            path = self._clustering_path
            self._kwik.write_attr(path, name, value)
        self._clustering_metadata.update(metadata)

    def _load_traces(self):
        # n_channels = self._metadata.get('n_channels', None)
        # dtype = self._metadata.get('dtype', None)
        # dtype = np.dtype(dtype) if dtype else None
        traces, opened_files = _read_traces(self._kwik,
                                            # dtype=dtype,
                                            # n_channels=n_channels,
                                            )

        if not traces:
            return

        # Update the list of opened files for cleanup.
        self._opened_files.extend(opened_files)

        # Set the recordings offsets (no delay between consecutive recordings).
        i = 0
        self._recording_offsets = []
        for trace in traces:
            self._recording_offsets.append(i)
            i += trace.shape[0] + 1
        self._traces = _concatenate_virtual_arrays(traces,
                                                   self._channel_order)
        self._all_traces = _concatenate_virtual_arrays(traces)

    def open(self, kwik_path, channel_group=None, clustering=None):
        """Open a Kwik dataset.

        The `.kwik` and `.kwx` must be in the same folder with the
        same basename.

        The files containing the traces (`.raw.kwd` or `.dat` / `.bin`) are
        determined according to the following logic:

        - Is there a path specified to a file which exists in
        [KWIK]/recordings/[X]/raw? If so, open it.
        - If this file does not exist, does a file exist with the same name
        in the current directory? If so, open it.
        - If such a file does not exist, or no filename is specified in
        the [KWIK], then is there a `raw.kwd` with the experiment basename
        in the current directory? If so, open it.
        - If not, return with a warning.

        Notes
        -----

        The `.kwik` file is opened in read-only mode, and is automatically
        closed when this function returns. It is temporarily reopened when
        the channel group or clustering changes.

        The `.kwik` file is temporarily opened in append mode when saving.

        The `.kwx` and `.raw.kwd` or `.dat` / `.bin` files stay open in
        read-only mode as long as `model.close()` is not called. This is
        because there might be read accesses to `features_masks` (`.kwx`)
        and waveforms (`.raw.kwd` or `.dat` / `.bin`) while the dataset is
        opened.

        Parameters
        ----------

        kwik_path : str
            Path to a `.kwik` file.
        channel_group : int or None (default is None)
            The channel group (shank) index to use. This can be changed
            later after the file has been opened. By default, the first
            channel group is used.
        clustering : str or None (default is None)
            The clustering to use. This can be changed later after the file
            has been opened. By default, the `main` clustering is used. An
            error is raised if the `main` clustering doesn't exist.

        """

        if kwik_path is None:
            raise ValueError("No kwik_path specified.")

        if not kwik_path.endswith('.kwik'):
            raise ValueError("File does not end in .kwik.")

        # Open the file.
        kwik_path = op.realpath(kwik_path)
        self.kwik_path = kwik_path
        self.name = op.splitext(op.basename(kwik_path))[0]

        # Open the KWIK file.
        self._kwik = _open_h5_if_exists(kwik_path, 'kwik')
        if self._kwik is None:
            raise IOError("File `{0}` doesn't exist.".format(kwik_path))
        if not self._kwik.is_open():
            raise IOError("File `{0}` failed to open.".format(kwik_path))
        self._check_kwik_version()

        # Load the metadata.
        self._metadata = self._load_meta('spikedetekt')
        self._kk2_metadata = self._load_meta('klustakwik2')

        # This needs the metadata.
        self._create_waveform_loader()

        self._load_recordings()

        # This generates the recording offset.
        self._load_traces()

        self._load_channel_groups(channel_group)

        # Load the probe.
        self._load_probe()

        # This needs channel groups.
        self._load_clusterings(clustering)

        # This needs the recording offsets.
        # This loads channels, channel_order, spikes, probe.
        self._channel_group_changed(self._channel_group)

        # This loads spike clusters and cluster groups.
        self._clustering_changed(clustering)

        # This needs channels, channel_order, and waveform loader.
        self._update_waveform_loader()

        # No need to keep the kwik file open.
        self._kwik.close()

        logger.debug("Kwik file `%s` loaded.", self.kwik_path)

    def save(self,
             spike_clusters=None,
             cluster_groups=None,
             clustering_metadata=None,
             ):
        """Save the spike clusters and cluster groups in the Kwik file."""

        # REFACTOR: with() to open/close the file if needed
        to_close = self._open_kwik_if_needed(mode='a')

        if spike_clusters is not None:
            self._save_spike_clusters(spike_clusters)
        if cluster_groups is not None:
            self._save_cluster_groups(cluster_groups)
        if clustering_metadata is not None:
            self._save_clustering_metadata(clustering_metadata)
        logger.info("Save the Kwik file at `%s`.", self.kwik_path)

        if to_close:
            self._kwik.close()

    def describe(self):
        """Display information about the dataset."""
        def _print(name, value):
            print("{0: <24}{1}".format(name, value))
        _print("Kwik file", self.kwik_path)
        _print("Recordings", self.n_recordings)

        # List of channel groups.
        cg = ['{:d}'.format(g) + ('*' if g == self.channel_group else '')
              for g in self.channel_groups]
        _print("List of shanks", ', '.join(cg))

        # List of clusterings.
        cl = ['{:s}'.format(c) + ('*' if c == self.clustering else '')
              for c in self.clusterings]
        _print("Clusterings", ', '.join(cl))

        _print("Channels", self.n_channels)
        _print("Spikes", self.n_spikes)
        _print("Clusters", self.n_clusters)
        _print("Duration", "{:.0f}s".format(self.duration))

    # Changing channel group and clustering
    # -------------------------------------------------------------------------

    def _channel_group_changed(self, value):
        """Called when the channel group changes."""
        if value not in self.channel_groups:
            raise ValueError("The channel group {0} is invalid.".format(value))
        self._channel_group = value

        # Load data.
        _to_close = self._open_kwik_if_needed()

        if self._kwik.h5py_file:
            clusterings = _list_clusterings(self._kwik.h5py_file,
                                            self._channel_group)
        else:
            logger.warn(".kwik filepath doesn't exist.")
            clusterings = None

        if clusterings:
            if 'main' in clusterings:
                self._load_clusterings('main')
                self.clustering = 'main'
            else:
                self._load_clusterings(clusterings[0])
                self.clustering = clusterings[0]
        else:
            self._spike_clusters = None
            self._clustering = None

        self._probe.change_channel_group(value)
        self._load_channels()
        self._load_spikes()
        self._load_features_masks()

        if _to_close:
            self._kwik.close()

        # Update the list of channels for the waveform loader.
        # self._waveform_loader.channels = self._channel_order

    def _clustering_changed(self, value):
        """Called when the clustering changes."""
        if value is None:
            return
        if value not in self.clusterings:
            raise ValueError("The clustering {0} is invalid.".format(value))
        self._clustering = value

        # Load data.
        _to_close = self._open_kwik_if_needed()
        self._create_cluster_metadata()
        self._load_spike_clusters()
        self._load_cluster_groups()
        self._load_clustering_metadata()
        if _to_close:
            self._kwik.close()

    # Managing cluster groups
    # -------------------------------------------------------------------------

    def _write_cluster_group(self, group_id, name, write_color=True):
        if group_id <= 3:
            raise ValueError("Default groups cannot be changed.")

        _DEFAULT_GROUPS.append((group_id, name))

        _to_close = self._open_kwik_if_needed(mode='a')

        _create_cluster_group(self._kwik, group_id, name,
                              clustering=self._clustering,
                              channel_group=self._channel_group,
                              write_color=write_color,
                              )

        if _to_close:
            self._kwik.close()

    def add_cluster_group(self, group_id, name):
        """Add a new cluster group."""
        self._write_cluster_group(group_id, name, write_color=True)

    def rename_cluster_group(self, group_id, name):
        """Rename an existing cluster group."""
        self._write_cluster_group(group_id, name, write_color=False)

    def delete_cluster_group(self, group_id):
        if group_id <= 3:
            raise ValueError("Default groups cannot be deleted.")

        path = ('/channel_groups/{0:d}/'
                'cluster_groups/{1:s}/{2:d}').format(self._channel_group,
                                                     self._clustering,
                                                     group_id,
                                                     )

        _to_close = self._open_kwik_if_needed(mode='a')

        self._kwik.delete(path)

        if _to_close:
            self._kwik.close()

    # Managing clusterings
    # -------------------------------------------------------------------------

    def add_clustering(self, name, spike_clusters):
        """Save a new clustering to the file."""
        if name in self._clusterings:
            raise ValueError("The clustering '{0}' ".format(name) +
                             "already exists.")
        assert len(spike_clusters) == self.n_spikes

        _to_close = self._open_kwik_if_needed(mode='a')

        logger.debug("Add clustering %s.", name)
        _create_clustering(self._kwik,
                           name,
                           channel_group=self._channel_group,
                           spike_clusters=spike_clusters,
                           )

        # Update the list of clusterings.
        self._load_clusterings(self._clustering)

        if _to_close:
            self._kwik.close()

    def _move_clustering(self, old_name, new_name, copy=None):
        if not copy and old_name == self._clustering:
            raise ValueError("You cannot move the current clustering.")
        if new_name in self._clusterings:
            raise ValueError("The clustering '{0}' ".format(new_name) +
                             "already exists.")

        _to_close = self._open_kwik_if_needed(mode='a')

        if copy:
            func = self._kwik.copy
        else:
            func = self._kwik.move

        # /channel_groups/x/spikes/clusters/<name>
        p = self._spikes_path + '/clusters/'
        func(p + old_name, p + new_name)

        # /channel_groups/x/clusters/<name>
        p = self._clusters_path + '/'
        func(p + old_name, p + new_name)

        # /channel_groups/x/cluster_groups/<name>
        p = self._channel_groups_path + '/cluster_groups/'
        func(p + old_name, p + new_name)

        # Update the list of clusterings.
        self._load_clusterings(self._clustering)

        if _to_close:
            self._kwik.close()

    def rename_clustering(self, old_name, new_name):
        """Rename a clustering in the `.kwik` file."""
        logger.debug("Rename clustering %s => %s.", old_name, new_name)
        self._move_clustering(old_name, new_name, copy=False)

    def copy_clustering(self, name, new_name):
        """Copy a clustering in the `.kwik` file."""
        logger.debug("Copy clustering %s => %s.", name, new_name)
        self._move_clustering(name, new_name, copy=True)

    def delete_clustering(self, name):
        """Delete a clustering."""
        if name == self._clustering:
            raise ValueError("You cannot delete the current clustering.")
        if name not in self._clusterings:
            raise ValueError(("The clustering {0} "
                              "doesn't exist.").format(name))

            logger.debug("Delete clustering %s.", name)
        _to_close = self._open_kwik_if_needed(mode='a')

        # /channel_groups/x/spikes/clusters/<name>
        parent = self._kwik.read(self._spikes_path + '/clusters/')
        del parent[name]

        # /channel_groups/x/clusters/<name>
        parent = self._kwik.read(self._clusters_path)
        del parent[name]

        # /channel_groups/x/cluster_groups/<name>
        parent = self._kwik.read(self._channel_groups_path +
                                 '/cluster_groups/')
        del parent[name]

        # Update the list of clusterings.
        self._load_clusterings(self._clustering)

        if _to_close:
            self._kwik.close()

    # Data
    # -------------------------------------------------------------------------

    @property
    def duration(self):
        """Duration of the experiment (in seconds)."""
        if self._traces is None:
            return 0.
        return float(self._traces.shape[0]) / self.sample_rate

    @property
    def channel_group(self):
        return self._channel_group

    @channel_group.setter
    def channel_group(self, value):
        assert isinstance(value, integer_types)
        self._channel_group = value
        self._channel_group_changed(value)

    @property
    def channel_groups(self):
        """List of channel groups found in the Kwik file."""
        return self._channel_groups

    @property
    def n_features_per_channel(self):
        """Number of features per channel (generally 3)."""
        return self._metadata['n_features_per_channel']

    @property
    def channels(self):
        """List of all channels in the current channel group.

        This list comes from the /channel_groups HDF5 group in the Kwik file.

        """
        # TODO: rename to channel_ids?
        return self._channels

    @property
    def channel_order(self):
        """List of kept channels in the current channel group.

        If you want the channels used in the features, masks, and waveforms,
        this is the property you want to use, and not `model.channels`.

        The channel order is the same than the one from the PRB file.
        This order was used when generating the features and masks
        in SpikeDetekt2. The same order is used in klusta when loading the
        waveforms from the traces file(s).

        """
        return self._channel_order

    @property
    def n_channels(self):
        """Number of all channels in the current channel group."""
        return len(self._channels)

    @property
    def recordings(self):
        """List of recording indices found in the Kwik file."""
        return self._recordings

    @property
    def n_recordings(self):
        """Number of recordings found in the Kwik file."""
        return len(self._recordings)

    @property
    def clusterings(self):
        """List of clusterings found in the Kwik file.

        The first one is always `main`.

        """
        return self._clusterings

    @property
    def clustering(self):
        """The currently-active clustering.

        Default is `main`.

        """
        return self._clustering

    @clustering.setter
    def clustering(self, value):
        """Change the currently-active clustering."""
        self._clustering_changed(value)

    @property
    def clustering_metadata(self):
        """A dictionary of key-value metadata specific to the current
        clustering."""
        return self._clustering_metadata

    @property
    def metadata(self):
        """A dictionary holding metadata about the experiment.

        This information comes from the PRM file. It was used by
        SpikeDetekt2 during spike detection.

        """
        return self._metadata

    @property
    def kk2_metadata(self):
        """A dictionary holding KK2 metadata."""
        return self._kk2_metadata

    @property
    def probe(self):
        """A `Probe` instance representing the probe used for the recording.

        This object contains information about the adjacency graph and
        the channel positions.

        """
        return self._probe

    @property
    def channel_positions(self):
        return self.probe.positions

    @property
    def all_traces(self):
        """Raw traces as found in the traces file(s).

        This object is memory-mapped to the HDF5 file, or `.dat` / `.bin` file,
        or both.

        """
        return self._all_traces

    @property
    def traces(self):
        """Like all_traces, but without the dead channels."""
        return self._traces

    @property
    def spike_samples(self):
        """Spike samples from the current channel group.

        This is a NumPy array containing `int64` values (number of samples
        in unit of the sample rate).

        The spike times of all recordings are concatenated. There is no gap
        between consecutive recordings, currently.

        """
        return self._spike_samples

    @property
    def sample_rate(self):
        """Sample rate of the recording.

        This value is found in the metadata coming from the PRM file.

        """
        return float(self._metadata['sample_rate'])

    @property
    def spike_recordings(self):
        """The recording index for each spike.

        This is a NumPy array of integers with `n_spikes` elements.

        """
        return self._spike_recordings

    @property
    def n_spikes(self):
        """Number of spikes in the current channel group."""
        return (len(self._spike_samples)
                if self._spike_samples is not None else 0)

    @property
    def all_features(self):
        """Features from the current channel group.

        This is memory-mapped to the `.kwx` file.

        Note: in general, it is better to use the cluster store to access
        the features and masks of some clusters.

        """
        return self._features

    @property
    def all_masks(self):
        """Masks from the current channel group.

        This is memory-mapped to the `.kwx` file.

        Note: in general, it is better to use the cluster store to access
        the features and masks of some clusters.

        """
        return self._masks

    @property
    def all_features_masks(self):
        """Features-masks from the current channel group.

        This is memory-mapped to the `.kwx` file.

        Note: in general, it is better to use the cluster store to access
        the features and masks of some clusters.

        """
        return self._features_masks

    @property
    def all_waveforms(self):
        """High-passed filtered waveforms from the current channel group.

        This is a virtual array mapped to the traces file(s). Filtering is
        done on the fly.

        The shape is `(n_spikes, n_samples, n_channels)`.

        """
        return SpikeLoader(self._waveform_loader, self.spike_samples)

    @property
    def spike_clusters(self):
        """Spike clusters from the current channel group and clustering.

        Every element is the cluster identifier of a spike.

        The shape is `(n_spikes,)`.

        """
        return self._spike_clusters

    @property
    def spike_times(self):
        """Spike times from the current channel_group.

        This is a NumPy array containing `float64` values (in seconds).

        The spike times of all recordings are concatenated. There is no gap
        between consecutive recordings, currently.

        """
        if self._spike_times is None:
            self._spike_times = (self.spike_samples.astype(np.float64) /
                                 self.sample_rate)
        return self._spike_times

    @property
    def cluster_metadata(self):
        """Metadata about the clusters in the current channel group and
        clustering.

        `cluster_metadata.group(cluster_id)` returns the group of a given
        cluster. The default group is 'unsorted'.

        """
        return self._cluster_metadata

    @property
    def cluster_groups(self):
        """Groups of all clusters in the current channel group and clustering.

        This is a regular Python dictionary.

        """
        return {cluster: self._cluster_metadata[cluster]
                for cluster in self.cluster_ids}

    @property
    def cluster_ids(self):
        """List of cluster ids from the current channel group and clustering.

        This is a sorted list of unique cluster ids as found in the current
        `spike_clusters` array.

        """
        return _unique(self._spike_clusters)

    @property
    def spike_ids(self):
        """List of spike ids."""
        return np.arange(self.n_spikes, dtype=np.int32)

    @property
    def n_clusters(self):
        """Number of clusters in the current channel group and clustering."""
        return len(self.cluster_ids)

    # Close
    # -------------------------------------------------------------------------

    def close(self):
        """Close the `.kwik` and `.kwx` files if they are open, and cleanup
        handles to all raw data files"""

        logger.debug("Closing files")
        del self._features_masks
        for f in self._opened_files:
            # upside-down if statement to avoid false positive lint error
            if not (isinstance(f, np.ndarray)):
                f.close()
            else:
                del f
        self._kwik.close()
